/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

#define DSL_INTERN

#include "drv_dsl_cpe_api.h"

#if defined(INCLUDE_DSL_PM)

#include "drv_dsl_cpe_pm_core.h"
#include "drv_dsl_cpe_device_pm.h"

#undef DSL_DBG_BLOCK
#define DSL_DBG_BLOCK DSL_DBG_PM

/** \addtogroup DRV_DSL_CPE_PM
 @{ */

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_Start(
   DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_Context *pPmContext = DSL_NULL;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: PM module starting..." DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if(pContext->PM != DSL_NULL)
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: WARNING - PM module already started" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_SUCCESS;
   }

   /* Create PM context */
   pPmContext = (DSL_PM_Context*)DSL_DRV_Malloc(sizeof(DSL_PM_Context));
   if(!pPmContext)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: PM_Start: no memory for internal context" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }
   memset(pPmContext, 0, sizeof(DSL_PM_Context));

   /* Initialize pointer to the PM context in the DSL CPE context*/
   pContext->PM = (DSL_void_t*)pPmContext;

   /* Create PM counters data*/
   DSL_DRV_PM_CONTEXT(pContext)->pCounters =
      (DSL_PM_CountersData_t*)DSL_DRV_Malloc(sizeof(DSL_PM_CountersData_t));
   if (DSL_DRV_PM_CONTEXT(pContext)->pCounters == DSL_NULL)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: PM_Start: no memory for PM counters!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }
   memset(DSL_DRV_PM_CONTEXT(pContext)->pCounters, 0x0, sizeof(DSL_PM_CountersData_t));

   /* Create PM counters dump data*/
   DSL_DRV_PM_CONTEXT(pContext)->pCountersDump =
      (DSL_PM_CountersDump_t*)DSL_DRV_Malloc(sizeof(DSL_PM_CountersDump_t));
   if (DSL_DRV_PM_CONTEXT(pContext)->pCountersDump == DSL_NULL)
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: PM_Start: no memory for PM dump counters!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }
   memset(DSL_DRV_PM_CONTEXT(pContext)->pCountersDump, 0x0, sizeof(DSL_PM_CountersDump_t));

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
   /* Set PM module line state information*/
   DSL_DRV_PM_CONTEXT(pContext)->nLineState     = DSL_LINESTATE_NOT_INITIALIZED;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/

   /* init PM module common mutex */
   DSL_DRV_MUTEX_INIT(DSL_DRV_PM_CONTEXT(pContext)->pmMutex);
   /* init PM module Near-End mutex */
   DSL_DRV_MUTEX_INIT(DSL_DRV_PM_CONTEXT(pContext)->pmNeMutex);
   /* init PM module Far-End mutex */
   DSL_DRV_MUTEX_INIT(DSL_DRV_PM_CONTEXT(pContext)->pmFeMutex);

   DSL_DRV_PM_CONTEXT(pContext)->nFeRequestCycle = DSL_PM_COUNTER_FE_POLLING_FACTOR;

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
   /* Set PM module default Sync Mode*/
   DSL_DRV_PM_CONTEXT(pContext)->syncMode = DSL_PM_DEFAULT_SYNC_MODE;

   /* Set PM module timeframes default values*/
   DSL_DRV_PM_CONTEXT(pContext)->nPm15Min = DSL_PM_15MIN;
   DSL_DRV_PM_CONTEXT(pContext)->nPm1Day  = DSL_PM_15MIN * DSL_PM_15MIN_PER_DAY;
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
   DSL_DRV_PM_CONTEXT(pContext)->nPmTick  = DSL_PM_COUNTER_POLLING_CYCLE;

#ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS
   #ifdef INCLUDE_DSL_CPE_PM_HISTORY
   /* Create Line Counters History*/
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.hist15min),
      DSL_PM_LINE_15MIN_RECORDS_NUM + 1);
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.hist1day),
      DSL_PM_LINE_1DAY_RECORDS_NUM + 1);

   #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
   #ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS
   /* Set default 15-min thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.thresholds15min),
      0xFF,
      sizeof(DSL_pmLineData_t));

   /* Set default 1-day thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.thresholds1day),
      0xFF,
      sizeof(DSL_pmLineData_t));
   #endif /* INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

   #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.histShowtime),
      DSL_PM_LINE_SHOWTIME_RECORDS_NUM + 1);
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS
   #ifdef INCLUDE_DSL_CPE_PM_HISTORY
   /* Create Channel Counters History*/
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.hist15min),
      DSL_PM_CHANNEL_15MIN_RECORDS_NUM + 1);
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.hist1day),
      DSL_PM_CHANNEL_1DAY_RECORDS_NUM + 1);

   #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
   #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS
   /* Set default 15-min thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.thresholds15min),
      0xFF,
      sizeof(DSL_pmChannelData_t));

   /* Set default 1-day thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.thresholds1day),
      0xFF,
      sizeof(DSL_pmChannelData_t));
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

   #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.histShowtime),
      DSL_PM_CHANNEL_SHOWTIME_RECORDS_NUM + 1);
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS
   #ifdef INCLUDE_DSL_CPE_PM_HISTORY
   /* Create Data Path Counters History*/
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.hist15min),
      DSL_PM_DATAPATH_15MIN_RECORDS_NUM + 1);
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.hist1day),
      DSL_PM_DATAPATH_1DAY_RECORDS_NUM + 1);

   #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
   #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS
   /* Set default 15-min thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.thresholds15min),
      0xFF,
      sizeof(DSL_pmDataPathData_t));

   /* Set default 1-day thresholds*/
   memset(
      &(DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.thresholds1day),
      0xFF,
      sizeof(DSL_pmDataPathData_t));
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

   #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
   DSL_DRV_PM_HistoryCreate(&(DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.histShowtime),
      DSL_PM_DATAPATH_SHOWTIME_RECORDS_NUM + 1);
   #endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS*/

   /* Initialize PM module device specific parameters*/
   if( DSL_DRV_PM_DEV_Start(pContext) != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM module device specific init failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   DSL_DRV_PM_CONTEXT(pContext)->bInit = DSL_TRUE;

   /*
      Init PM module threads
   */
   DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.bRun = DSL_FALSE;
   DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.nThreadPollTime = DSL_PM_COUNTER_POLLING_CYCLE;
   DSL_DRV_INIT_EVENT("pmev_ne", DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.pmEvent);

   /* Start PM module Near-End thread*/
   nErrCode = DSL_DRV_THREAD(&DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.Control, "pmex_ne", DSL_DRV_PM_ThreadNe, (int)pContext);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - PM module NE thread start failed, retCode(%d)!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), nErrCode));
   }

   DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.bRun = DSL_FALSE;
   DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.nThreadPollTime =
      DSL_PM_COUNTER_POLLING_CYCLE * DSL_DRV_PM_CONTEXT(pContext)->nFeRequestCycle;
   DSL_DRV_INIT_EVENT("pmev_fe", DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.pmEvent);

   /* Start PM module Far-End thread*/
   nErrCode = DSL_DRV_THREAD(&DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.Control, "pmex_fe", DSL_DRV_PM_ThreadFe, (int)pContext);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - PM module FE thread start failed, retCode(%d)!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), nErrCode));
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: PM module started..." DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_Stop(
   DSL_Context_t *pContext)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_CTX_POINTER(pContext);
   DSL_CHECK_ERR_CODE();

   if( !DSL_DRV_PM_CONTEXT(pContext) )
   {
      return DSL_SUCCESS;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: Stopping PM module..."DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   if( DSL_DRV_PM_CONTEXT(pContext) == DSL_NULL || DSL_DRV_PM_CONTEXT(pContext)->bInit == DSL_FALSE )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: No PM context found or PM module was not initialized"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_SUCCESS;
   }

   /* Check bPmLock flag*/
   if( DSL_DRV_PM_CONTEXT(pContext)->bPmLock )
   {
      /*... Unlock if locked to allow PM module stop*/
      DSL_DRV_PM_UnLock(pContext);
   }

   /* Check the PM module Far-End thread active flag*/
   if( DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.bRun != DSL_TRUE )
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: PM module Near-End thread already stopped"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
   else
   {
      /* Signal PM Far-End thread to stop*/
      DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.bRun = DSL_FALSE;

      DSL_DRV_WAKEUP_EVENT(DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.pmEvent);
      DSL_DRV_WAIT_COMPLETION(&DSL_DRV_PM_CONTEXT(pContext)->pmThreadFe.Control);

      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: PM Far-End thread has stopped... (%lu)"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), DSL_DRV_TimeMSecGet()));
   }

   /* Check the PM module Near-End thread active flag*/
   if( DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.bRun != DSL_TRUE )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: PM module Near-End thread already stopped"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
   else
   {
      /* Signal PM Near-End thread to stop*/
      DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.bRun = DSL_FALSE;

      DSL_DRV_WAKEUP_EVENT(DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.pmEvent);
      DSL_DRV_WAIT_COMPLETION(&DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.Control);

      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: PM Near-End thread has stopped... (%lu)"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), DSL_DRV_TimeMSecGet()));
   }

   if (DSL_DRV_PM_CONTEXT(pContext)->pCounters != DSL_NULL)
   {
      DSL_DRV_MemFree(DSL_DRV_PM_CONTEXT(pContext)->pCounters);
   }

   if (DSL_DRV_PM_CONTEXT(pContext)->pCountersDump != DSL_NULL)
   {
      DSL_DRV_MemFree(DSL_DRV_PM_CONTEXT(pContext)->pCountersDump);
   }

   /* Free PM Module resources*/
   if( pContext->PM != DSL_NULL )
   {
      DSL_DRV_MemFree(pContext->PM);
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: PM module has stopped... (%lu)"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), DSL_DRV_TimeMSecGet()));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_Reset(
   DSL_Context_t *pContext,
   DSL_PM_Reset_t *pResetType)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pResetType);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_PM_Reset"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module processing*/
   nErrCode = DSL_DRV_PM_Lock(pContext);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - PM module lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS
   /* Reset Channel Counters*/
   nErrCode = DSL_DRV_PM_ChannelCountersReset(pContext, pResetType->data.nResetType);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - Channel counters reset failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
#endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS
   /* Reset Line Counters*/
   nErrCode = DSL_DRV_PM_LineCountersReset(pContext, pResetType->data.nResetType);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - Line counters reset failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS
   /* Reset Data Path Counters*/
   nErrCode = DSL_DRV_PM_DataPathCountersReset(pContext, pResetType->data.nResetType);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG(DSL_DBG_MSG,
         (pContext, "DSL[%02d]: ERROR - Data Path counters reset failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS*/

   /* Unlock PM module processing*/
   DSL_DRV_PM_UnLock(pContext);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_PM_Reset"DSL_DRV_CRLF, DSL_DEV_NUM(pContext)));

   return nErrCode;
}

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description please refer to the equivalent ioctl
   \ref DSL_FIO_PM_15MIN_ELAPSED_EXT_TRIGGER
*/
DSL_uint32_t DSL_DRV_PM_15MinElapsedExtTrigger(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_ElapsedExtTrigger_t *pTrigger)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_boolean_t b15min = DSL_FALSE, b1day = DSL_FALSE;

   DSL_CHECK_POINTER(pContext, pTrigger);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_15MinElapsedExtTrigger" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if (pTrigger->data.bOneDayElapsed)
   {
      /* Trigger both 15-min and 1-day intervals*/
      b15min = b1day = DSL_TRUE;
   }
   else
   {
      /* Trigger 15-min interval*/
      b15min = DSL_TRUE;
   }

   /* Lock PM module Mutex*/
   if( DSL_DRV_MUTEX_LOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex) )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Couldn't lock PM mutex!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERR_SEMAPHORE_GET;
   }

   /* Check for external 1-day elapsed trigger*/
   if( DSL_DRV_PM_CONTEXT(pContext)->bExternal1dayElapsed == DSL_FALSE )
   {
      if (b1day)
      {
         DSL_DRV_PM_CONTEXT(pContext)->bExternal1dayElapsed = DSL_TRUE;
      }
   }
   else
   {
      DSL_DEBUG( DSL_DBG_WRN,
         (pContext, "DSL[%02d]: Previous 1-day elapsed trigger not proceeded yet"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      nErrCode = DSL_WRN_PM_PREVIOUS_EXTERNAL_TRIGGER_NOT_HANDLED;
   }

   /* Check for external 15-min elapsed trigger*/
   if( DSL_DRV_PM_CONTEXT(pContext)->bExternal15minElapsed == DSL_FALSE )
   {
      if (b15min)
      {
         DSL_DRV_PM_CONTEXT(pContext)->bExternal15minElapsed = DSL_TRUE;
      }
   }
   else
   {
      DSL_DEBUG( DSL_DBG_WRN,
         (pContext, "DSL[%02d]: Previous 15-min elapsed trigger not proceeded yet"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      nErrCode = DSL_WRN_PM_PREVIOUS_EXTERNAL_TRIGGER_NOT_HANDLED;
   }

   /* Unlock PM module Mutex*/
   DSL_DRV_MUTEX_UNLOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_15MinElapsedExtTrigger, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return(nErrCode);
}


/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ElapsedTimeReset(
   DSL_Context_t *pContext,
   DSL_PM_ElapsedTimeReset_t *pReset)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pReset);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ElapsedTimeReset"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module Mutex*/
   if( DSL_DRV_MUTEX_LOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex) )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Couldn't lock PM mutex!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERR_SEMAPHORE_GET;
   }

   if (DSL_DRV_PM_CONTEXT(pContext)->syncMode != DSL_PM_SYNC_MODE_SYS_TIME)
   {
      /* Reset PM module elapsed time*/
      DSL_DRV_PM_CONTEXT(pContext)->nCurr15MinTime =
      DSL_DRV_PM_CONTEXT(pContext)->nCurr1DayTime  = 0;
   }
   else
   {
      nErrCode = DSL_WRN_PM_NOT_ALLOWED_IN_CURRENT_SYNC_MODE;
   }

   /* Unlock PM module Mutex*/
   DSL_DRV_MUTEX_UNLOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ElapsedTimeReset, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_SyncModeSet(
   DSL_Context_t *pContext,
   DSL_PM_SyncMode_t *pMode)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_SyncModeType_t mode;
   DSL_DRV_TimeVal_t currTime;

   DSL_CHECK_POINTER(pContext, pMode);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_SyncModeSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Check for default Sync Mode*/
   mode = pMode == DSL_NULL ? DSL_PM_DEFAULT_SYNC_MODE : pMode->data.nMode;

   if ( mode == DSL_DRV_PM_CONTEXT(pContext)->syncMode )
   {
      return DSL_SUCCESS;
   }

   /* Lock PM module Mutex*/
   if( DSL_DRV_MUTEX_LOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex) )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Couldn't lock PM mutex!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERR_SEMAPHORE_GET;
   }

   switch(mode)
   {
   case DSL_PM_SYNC_MODE_FREE:
   case DSL_PM_SYNC_MODE_EXTERNAL:
      /* Reset current 15 min and 1 day time*/
      DSL_DRV_PM_CONTEXT(pContext)->nCurr15MinTime = 0;
      DSL_DRV_PM_CONTEXT(pContext)->nCurr1DayTime  = 0;
   break;
   case DSL_PM_SYNC_MODE_SYS_TIME:
      currTime = DSL_DRV_PM_SYS_TIME_GET();
      /* Sync to System time*/
      DSL_DRV_PM_CONTEXT(pContext)->nCurr15MinTime = currTime % (DSL_DRV_PM_CONTEXT(pContext)->nPm15Min);
      DSL_DRV_PM_CONTEXT(pContext)->nCurr1DayTime  = currTime % (DSL_DRV_PM_CONTEXT(pContext)->nPm1Day);
   break;

   default:
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Unknown Sync Mode (%d)"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), mode));
      nErrCode = DSL_ERROR;
   break;
   }

   if( nErrCode == DSL_SUCCESS )
   {
      /* Set new Sync mode in the PM module context*/
      DSL_DRV_PM_CONTEXT(pContext)->syncMode = mode;
   }

   /* Unlock PM module Mutex*/
   DSL_DRV_MUTEX_UNLOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_SyncModeSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_BurninModeSet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_BurninMode_t *pBurninMode)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pBurninMode);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_BurninModeSet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if( pBurninMode->data.bActivate )
   {
      if( pBurninMode->data.nMode.nPmTick <  DSL_PM_COUNTER_MIN_POLLING_CYCLE ||
          pBurninMode->data.nMode.nPmTick == 0 ||
          pBurninMode->data.nMode.nPm15Min == 0 ||
          pBurninMode->data.nMode.nPm15MinPerDay == 0 )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - DSL_PM_BurninModeSet: invalid mode!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
         return DSL_ERR_INVALID_PARAMETR;
      }
   
      if( pBurninMode->data.nMode.nPmTick > (pBurninMode->data.nMode.nPm15Min*1000) )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - nPmTick can't be greater than nPm15Min!"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
         return DSL_ERR_INVALID_PARAMETR;
      }
   }

   /* Lock PM module Mutex*/
   if( DSL_DRV_MUTEX_LOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex) )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Couldn't lock PM mutex!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERR_SEMAPHORE_GET;
   }

   if( pBurninMode->data.bActivate )
   {

      /* Set new timeframes values*/
      DSL_DRV_PM_CONTEXT(pContext)->nPm15Min = pBurninMode->data.nMode.nPm15Min;
      DSL_DRV_PM_CONTEXT(pContext)->nPm1Day  =
         pBurninMode->data.nMode.nPm15Min * pBurninMode->data.nMode.nPm15MinPerDay;
      DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.nThreadPollTime =
         DSL_DRV_PM_CONTEXT(pContext)->nPmTick = pBurninMode->data.nMode.nPmTick;
   }
   else
   {
      /* restore default settings*/
      DSL_DRV_PM_CONTEXT(pContext)->nPm15Min = DSL_PM_15MIN;
      DSL_DRV_PM_CONTEXT(pContext)->nPm1Day  = DSL_PM_15MIN * DSL_PM_15MIN_PER_DAY;
      DSL_DRV_PM_CONTEXT(pContext)->pmThreadNe.nThreadPollTime =
         DSL_DRV_PM_CONTEXT(pContext)->nPmTick = DSL_PM_COUNTER_POLLING_CYCLE;
   }

   /* Unlock PM module Mutex*/
   DSL_DRV_MUTEX_UNLOCK(DSL_DRV_PM_CONTEXT(pContext)->pmMutex);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_BurninModeSet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

/* ************************************************************************** */
/* *  xDSL Channel Endpoint interface (internal)                            * */
/* ************************************************************************** */
#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelHistoryStats15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsChDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_DIRECTION(pStats->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_PM_ChannelHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_15MIN(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelHistoryStats1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsChDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_DIRECTION(pStats->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_1DAY(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelCounters15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_ChannelData_t *pChCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get History Fill Level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_15MIN(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get History Item Index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_15MIN(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index error!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 15min interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm15Min;
   }
   else
   {
      /* Get 15-min interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed15MinTime;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pChCounters = DSL_DRV_PM_PTR_CHANNEL_COUNTERS_15MIN(histIdx,pCounters->nChannel,pCounters->nDirection);

   if (pChCounters != DSL_NULL)
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pChCounters, sizeof(DSL_PM_ChannelData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelCounters15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelCounters1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelCounters_t   *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_ChannelData_t *pChCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelCounters1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get History Fill Level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_1DAY(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get History Item Index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_1DAY(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index error!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 1day interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm1Day;
   }
   else
   {
      /* Get 1-day interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed1DayTime;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pChCounters = DSL_DRV_PM_PTR_CHANNEL_COUNTERS_1DAY(histIdx,pCounters->nChannel,pCounters->nDirection);

   if (pChCounters != DSL_NULL)
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pChCounters, sizeof(DSL_PM_ChannelData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelCounters1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelCountersTotalGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelCountersTotal_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_ChannelData_t *pChCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelCountersTotalGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Fill Total Counters elapsed time*/
   pCounters->total.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPmTotalElapsedTime/DSL_PM_MSEC;

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pChCounters = DSL_DRV_PM_PTR_CHANNEL_COUNTERS_TOTAL(pCounters->nChannel,pCounters->nDirection);

   if (pChCounters != DSL_NULL)
   {
      /* Copy Total Channel Counters data*/
      memcpy(&(pCounters->data), pChCounters, sizeof(DSL_PM_ChannelData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - total counters data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      nErrCode = DSL_ERR_INTERNAL;
   }


   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelCountersTotalGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelThresholds15MinSet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_ChannelData_t *pChCounters, *pThresholds;
   DSL_pmChannelThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelThresholds15MinSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_15MIN(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history item index get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pChCounters = DSL_DRV_PM_PTR_CHANNEL_COUNTERS_15MIN(histIdx,pThreshs->nChannel,pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_IND(pThreshs->nDirection);
   pInd = &(pInd[pThreshs->nChannel]);
   pThresholds = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_15MIN(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   /* Update Channel thresholds*/
   nErrCode = DSL_DRV_PM_ChannelThresholdsUpdate(
                 pChCounters, pThresholds, &(pThreshs->data), &(pInd->n15Min));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Channel 15-min thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelThresholds15MinSet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelThresholds1DaySet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_ChannelData_t *pChCounters, *pThresholds;
   DSL_pmChannelThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelThresholds1DaySet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_1DAY(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history item index get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pChCounters = DSL_DRV_PM_PTR_CHANNEL_COUNTERS_1DAY(histIdx,pThreshs->nChannel,pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_IND(pThreshs->nDirection);
   pInd = &(pInd[pThreshs->nChannel]);

   pThresholds = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_1DAY(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   /* Update Channel thresholds*/
   nErrCode = DSL_DRV_PM_ChannelThresholdsUpdate(
                 pChCounters, pThresholds, &(pThreshs->data), &(pInd->n1Day));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Channel 1-day thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelThresholds1DaySet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelThresholds15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_ChannelData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelThresholds15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_15MIN(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_ChannelData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Channel 15-min thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelThresholds15MinGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelThresholds1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_ChannelData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelThresholds1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      
      return nErrCode;
   }


   pThresholds = DSL_DRV_PM_PTR_CHANNEL_THRESHOLD_1DAY(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_ChannelData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Channel 1-day thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_ChannelThresholds1DayGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_ChannelCountersShowtimeGet(
   DSL_Context_t *pContext,
   DSL_PM_ChannelCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0, histInv = 0;
   DSL_PM_ChannelData_t *pChCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_ChannelCountersShowtimeGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if( DSL_DRV_PM_CONTEXT(pContext)->bShowtimeProcessingStart == DSL_FALSE )
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: No showtime data for the specified interval"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      memset( &(pCounters->data), 0x0, sizeof(DSL_PM_ChannelData_t));
      pCounters->interval.bValid       = DSL_FALSE;
      pCounters->interval.nElapsedTime = 0;
      pCounters->interval.nNumber      = 0;

      return DSL_WRN_PM_NO_SHOWTIME_DATA;
   }

   /* Clear the output structure*/
   pCounters->interval.nElapsedTime = 0;
   memset(&(pCounters->data), 0x0, sizeof(DSL_PM_ChannelData_t));

   /* Get History Fill Level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_SHOWTIME(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));
      
      return nErrCode;
   }

   for( histInv = 0; histInv <= pCounters->nHistoryInterval; histInv++ )
   {
      /* Get History Item Index*/
      nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                    pContext, DSL_DRV_PM_PTR_CHANNEL_HISTORY_SHOWTIME(),
                    histInv, &histIdx);
      if( nErrCode != DSL_SUCCESS || histIdx < 0 )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history index error!" DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         break;
      }

      /* Set Showtime Elapsed time*/
      pCounters->interval.nElapsedTime +=
         DSL_DRV_PM_CONTEXT(pContext)->pCounters->channelCounters.nShowtimeTimeHist[histIdx];

      pChCounters =
         DSL_DRV_PM_PTR_CHANNEL_COUNTERS_SHOWTIME(histIdx,pCounters->nChannel,pCounters->nDirection);

      if (pChCounters != DSL_NULL)
      {
         pCounters->data.nFEC            += pChCounters->nFEC;
         pCounters->data.nCodeViolations += pChCounters->nCodeViolations;
      }
      else
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         nErrCode = DSL_ERR_INTERNAL;
         
         break;
      }
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if( nErrCode == DSL_SUCCESS )
   {
      pCounters->interval.bValid  = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
   }

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG, (pContext,
      "DSL[%02d]: OUT - DSL_DRV_PM_ChannelCountersShowtimeGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_CHANNEL_COUNTERS*/

/* ************************************************************************** */
/* *  xDSL Data-Path Endpoint interface (internal)                          * */
/* ************************************************************************** */
#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathHistoryStats15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsChDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_DIRECTION(pStats->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_15MIN(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathHistoryStats1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsChDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathHistoryStats1DayGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_1DAY(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathHistoryStats1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathCounters15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_DataPathData_t *pDpCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_15MIN(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
         
      return DSL_ERROR;
   }

   /* Get history item index for the specified interval*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_15MIN(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 15min interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm15Min;
   }
   else
   {
      /* Get 15-min interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed15MinTime;
   }

   /* Lock PM module dircetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      
      return nErrCode;
   }

   pDpCounters = DSL_DRV_PM_PTR_DATAPATH_COUNTERS_15MIN(histIdx,pCounters->nChannel,pCounters->nDirection);

   if( pDpCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pDpCounters, sizeof(DSL_PM_DataPathData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathCounters15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathCounters1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_DataPathData_t *pDpCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathCounters1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_1DAY(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get history item index for the specified interval*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_1DAY(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 1day interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm1Day;
   }
   else
   {
      /* Get 1-day interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed1DayTime;
   }

   /* Lock PM module dircetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pDpCounters = DSL_DRV_PM_PTR_DATAPATH_COUNTERS_1DAY(histIdx,pCounters->nChannel,pCounters->nDirection);

   if( pDpCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pDpCounters, sizeof(DSL_PM_DataPathData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathCounters1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathCountersTotalGet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathCountersTotal_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_DataPathData_t *pDpCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathCountersTotalGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Fill Total Counters elapsed time*/
   pCounters->total.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPmTotalElapsedTime/DSL_PM_MSEC;

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pDpCounters = DSL_DRV_PM_PTR_DATAPATH_COUNTERS_TOTAL(pCounters->nChannel,pCounters->nDirection);

   if (pDpCounters != DSL_NULL)
   {
      /* Copy Total Channel Counters data*/
      memcpy(&(pCounters->data), pDpCounters, sizeof(DSL_PM_DataPathData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - total counters data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathCountersTotalGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathThresholds15MinSet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_DataPathData_t *pDpCounters, *pThresholds;
   DSL_pmDataPathThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathThresholds15MinSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_15MIN(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pDpCounters = DSL_DRV_PM_PTR_DATAPATH_COUNTERS_15MIN(histIdx,pThreshs->nChannel,pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_IND(pThreshs->nDirection);
   pInd = &(pInd[pThreshs->nChannel]);
   pThresholds = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_15MIN(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_DataPathThresholdsUpdate(
                 pDpCounters, pThresholds, &(pThreshs->data), &(pInd->n15Min));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Data Path 15-min thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathThresholds15MinSet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathThresholds1DaySet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_DataPathData_t *pDpCounters, *pThresholds;
   DSL_pmDataPathThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathThresholds1DaySet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_1DAY(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      
      return nErrCode;
   }

   pDpCounters = DSL_DRV_PM_PTR_DATAPATH_COUNTERS_1DAY(histIdx,pThreshs->nChannel,pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_IND(pThreshs->nDirection);
   pInd = &(pInd[pThreshs->nChannel]);
   pThresholds = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_1DAY(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_DataPathThresholdsUpdate(
                 pDpCounters, pThresholds, &(pThreshs->data), &(pInd->n1Day));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Data Path 1-day thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathThresholds1DaySet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathThresholds15MinGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_DataPathThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_DataPathData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathThresholds15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_15MIN(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_DataPathData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Data Path 15-min thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathThresholds15MinGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathThresholds1DayGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_DataPathThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_DataPathData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_CHANNEL_RANGE(pThreshs->nChannel);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathThresholds1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_DATAPATH_THRESHOLD_1DAY(pThreshs->nDirection);
   pThresholds = &(pThresholds[pThreshs->nChannel]);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_DataPathData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Data Path 1-day thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_DataPathThresholds1DayGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_DataPathCountersShowtimeGet(
   DSL_Context_t *pContext,
   DSL_PM_DataPathCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0, histInv = 0;
   DSL_PM_DataPathData_t *pDpCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_CHANNEL_RANGE(pCounters->nChannel);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_DataPathCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if( DSL_DRV_PM_CONTEXT(pContext)->bShowtimeProcessingStart == DSL_FALSE )
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: No showtime data for the specified interval"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      memset( &(pCounters->data), 0x0, sizeof(DSL_PM_DataPathData_t));
      pCounters->interval.bValid       = DSL_FALSE;
      pCounters->interval.nElapsedTime = 0;
      pCounters->interval.nNumber      = 0;

      return DSL_WRN_PM_NO_SHOWTIME_DATA;
   }

   /* Clear the output structure*/
   pCounters->interval.nElapsedTime = 0;
   memset( &(pCounters->data), 0x0, sizeof(DSL_PM_DataPathData_t));

   /* Get History Fill Level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_SHOWTIME(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Lock PM module dircetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   for( histInv = 0; histInv <= pCounters->nHistoryInterval; histInv++ )
   {
      /* Get history item index for the specified interval*/
      nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                    pContext, DSL_DRV_PM_PTR_DATAPATH_HISTORY_SHOWTIME(),
                    histInv, &histIdx);
      if( nErrCode != DSL_SUCCESS || histIdx < 0 )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         break;
      }

      /* Set Showtime Elapsed time*/
      pCounters->interval.nElapsedTime +=
         DSL_DRV_PM_CONTEXT(pContext)->pCounters->dataPathCounters.nShowtimeTimeHist[histIdx];

      pDpCounters =
         DSL_DRV_PM_PTR_DATAPATH_COUNTERS_SHOWTIME(histIdx,pCounters->nChannel,pCounters->nDirection);

      if( pDpCounters != DSL_NULL )
      {
         pCounters->data.nCRC_P          += pDpCounters->nCRC_P;
         pCounters->data.nCRCP_P         += pDpCounters->nCRCP_P;
         pCounters->data.nCV_P           += pDpCounters->nCV_P;
         pCounters->data.nCVP_P          += pDpCounters->nCVP_P;
         pCounters->data.nHEC            += pDpCounters->nHEC;
         pCounters->data.nIBE            += pDpCounters->nIBE;
         pCounters->data.nTotalCells     += pDpCounters->nTotalCells;
         pCounters->data.nUserTotalCells += pDpCounters->nUserTotalCells;
      }
      else
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         nErrCode = DSL_ERR_INTERNAL;
         
         break;
      }
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if( nErrCode == DSL_SUCCESS )
   {
      pCounters->interval.bValid  = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
   }

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG, (pContext,
      "DSL[%02d]: OUT - DSL_DRV_PM_DataPathCountersShowtimeGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_DATA_PATH_COUNTERS*/

/* ************************************************************************** */
/* *  xDSL Line Endpoint interface (internal)                               * */
/* ************************************************************************** */
#ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecHistoryStats15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_DIRECTION(pStats->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecHistoryStats1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStatsDir_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_DIRECTION(pStats->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecCounters15MinGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineSecCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_LineSecData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 15min interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm15Min;
   }
   else
   {
      /* Get 15-min interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed15MinTime;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineCounters = DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_15MIN(histIdx, pCounters->nDirection);

   if( pLineCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pLineCounters, sizeof(DSL_PM_LineSecData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecCounters15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecCounters1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_LineSecCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_LineSecData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_ERR,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecCounters1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 15min interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm1Day;
   }
   else
   {
      /* Get 1-day interval elapsed time and convert msec to sec*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed1DayTime;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineCounters = DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_1DAY(histIdx, pCounters->nDirection);

   if( pLineCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pLineCounters, sizeof(DSL_PM_LineSecData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecCounters1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

#ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecThresholds15MinSet(
   DSL_Context_t *pContext,
   DSL_PM_LineSecThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_LineSecData_t *pLineSecCounters, *pThresholds;
   DSL_pmLineThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecThresholds15MinSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineSecCounters = DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_15MIN(histIdx, pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_LINE_THRESHOLD_IND(pThreshs->nDirection);
   pThresholds = DSL_DRV_PM_PTR_LINE_SEC_THRESHOLD_15MIN(pThreshs->nDirection);

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_LineSecThresholdsUpdate(
                 pLineSecCounters, pThresholds, &(pThreshs->data), &(pInd->n15Min));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Sec 15-min thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecThresholds15MinSet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecThresholds1DaySet(
   DSL_Context_t *pContext,
   DSL_PM_LineSecThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_LineSecData_t *pLineSecCounters, *pThresholds;
   DSL_pmLineThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecThresholds1DaySet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineSecCounters = DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_1DAY(histIdx, pThreshs->nDirection);

   pInd = DSL_DRV_PM_PTR_LINE_THRESHOLD_IND(pThreshs->nDirection);
   pThresholds = DSL_DRV_PM_PTR_LINE_SEC_THRESHOLD_1DAY(pThreshs->nDirection);

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_LineSecThresholdsUpdate(
                 pLineSecCounters, pThresholds, &(pThreshs->data), &(pInd->n1Day));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Sec 1-day thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecThresholds1DaySet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecThresholds15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_LineSecThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineSecData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecThresholds15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_LINE_SEC_THRESHOLD_15MIN(pThreshs->nDirection);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_LineSecData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Sec 15-min thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecThresholds15MinGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecThresholds1DayGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineSecThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineSecData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ATU_DIRECTION(pThreshs->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecThresholds1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_LINE_SEC_THRESHOLD_1DAY(pThreshs->nDirection);

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_LineSecData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Sec 1-day thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pThreshs->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecThresholds1DayGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecCountersTotalGet(
   DSL_Context_t *pContext,
   DSL_PM_LineSecCountersTotal_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineSecData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecCountersTotalGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Fill Total Counters elapsed time*/
   pCounters->total.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPmTotalElapsedTime/DSL_PM_MSEC;

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineCounters = DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_TOTAL(pCounters->nDirection);

   if (pLineCounters != DSL_NULL)
   {
      /* Copy Total Channel Counters data*/
      memcpy(&(pCounters->data), pLineCounters, sizeof(DSL_PM_LineSecData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - total counters data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineSecCountersTotalGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineSecCountersShowtimeGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineSecCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0, histInv = 0;
   DSL_PM_LineSecData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ATU_DIRECTION(pCounters->nDirection);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineSecCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if( DSL_DRV_PM_CONTEXT(pContext)->bShowtimeProcessingStart == DSL_FALSE )
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: No showtime data for the specified interval"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      memset( &(pCounters->data), 0x0, sizeof(DSL_PM_LineSecData_t));
      pCounters->interval.bValid       = DSL_FALSE;
      pCounters->interval.nElapsedTime = 0;
      pCounters->interval.nNumber      = 0;

      return DSL_WRN_PM_NO_SHOWTIME_DATA;
   }

   /* Clear the output structure*/
   pCounters->interval.nElapsedTime = 0;
   memset(&(pCounters->data), 0x0, sizeof(DSL_PM_LineSecData_t));

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_SHOWTIME(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   for( histInv = 0; histInv <= pCounters->nHistoryInterval; histInv++ )
   {
      /* Get history item index*/
      nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                    pContext, DSL_DRV_PM_PTR_LINE_HISTORY_SHOWTIME(),
                    histInv, &histIdx);
      if( nErrCode != DSL_SUCCESS || histIdx < 0 )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         break;
      }

      /* Set Showtime Elapsed time*/
      pCounters->interval.nElapsedTime +=
         DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.nShowtimeTimeHist[histIdx];

      pLineCounters =
         DSL_DRV_PM_PTR_LINE_SEC_COUNTERS_SHOWTIME(histIdx, pCounters->nDirection);

      if( pLineCounters != DSL_NULL )
      {
         pCounters->data.nES   += pLineCounters->nES;
         pCounters->data.nLOFS += pLineCounters->nLOFS;
         pCounters->data.nLOSS += pLineCounters->nLOSS;
         pCounters->data.nSES  += pLineCounters->nSES;
         pCounters->data.nUAS  += pLineCounters->nUAS;
      }
      else
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         nErrCode = DSL_ERR_INTERNAL;
         
         break;
      }
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, pCounters->nDirection, DSL_FALSE);

   if( nErrCode == DSL_SUCCESS )
   {
      pCounters->interval.bValid  = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
   }

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG, (pContext,
      "DSL[%02d]: OUT - DSL_DRV_PM_LineSecCountersShowtimeGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_HISTORY
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitHistoryStats15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStats_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitHistoryStats15MinGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitHistoryStats15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitHistoryStats1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_HistoryStats_t *pStats)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   DSL_CHECK_POINTER(pContext, pStats);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitHistoryStats1DayGet"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get PM history statistics information*/
   nErrCode = DSL_DRV_PM_HistoryStatsGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), &(pStats->data));

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG(DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitHistoryStats1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitCounters15MinGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineInitCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_LineInitData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 15min interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm15Min;
   }
   else
   {
      /* Get 15-min interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed15MinTime;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_15MIN(histIdx);

   if( pLineCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pLineCounters, sizeof(DSL_PM_LineInitData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitCounters15MinGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitCounters1DayGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineInitCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0;
   DSL_PM_LineInitData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitCounters1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
         
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(),
                 pCounters->nHistoryInterval, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0 )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   if( pCounters->nHistoryInterval != 0 )
   {
      /* Set Elapsed time to the number of sec in the 1day interval*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPm1Day;
   }
   else
   {
      /* Get 15-min interval elapsed time*/
      pCounters->interval.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nElapsed1DayTime;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pLineCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_1DAY(histIdx);

   if( pLineCounters != DSL_NULL )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
      memcpy(&(pCounters->data), pLineCounters, sizeof(DSL_PM_LineInitData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitCounters1DayGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

#ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS
#ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitThresholds15MinSet(
   DSL_Context_t *pContext,
   DSL_PM_LineInitThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_LineInitData_t *pLineInitCounters, *pThresholds;
   DSL_pmLineThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitThresholds15MinSet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_15MIN(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLineInitCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_15MIN(histIdx);

   pInd = DSL_DRV_PM_PTR_LINE_THRESHOLD_IND(DSL_NEAR_END);
   pThresholds = DSL_DRV_PM_PTR_LINE_INIT_THRESHOLD_15MIN();

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_LineInitThresholdsUpdate(
                 pLineInitCounters, pThresholds, &(pThreshs->data), &(pInd->n15Min));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Init 15-min thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitThresholds15MinSet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitThresholds1DaySet(
   DSL_Context_t *pContext,
   DSL_PM_LineInitThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_PM_LineInitData_t *pLineInitCounters, *pThresholds;
   DSL_pmLineThresholdCrossingData_t *pInd ;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitThresholds1DaySet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Get history item index*/
   nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_1DAY(), 0, &histIdx);

   if( nErrCode != DSL_SUCCESS || histIdx < 0)
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
      return nErrCode;
   }

   pLineInitCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_1DAY(histIdx);

   pInd = DSL_DRV_PM_PTR_LINE_THRESHOLD_IND(DSL_NEAR_END);
   pThresholds = DSL_DRV_PM_PTR_LINE_INIT_THRESHOLD_1DAY();

   /* Update Data Path thresholds*/
   nErrCode = DSL_DRV_PM_LineInitThresholdsUpdate(
                 pLineInitCounters, pThresholds, &(pThreshs->data), &(pInd->n1Day));

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Init 1-day thresholds update failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitThresholds1DaySet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitThresholds15MinGet(
   DSL_Context_t *pContext,
   DSL_PM_LineInitThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineInitData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitThresholds15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_LINE_INIT_THRESHOLD_15MIN();

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_LineInitData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Init 15-min thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitThresholds15MinGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}

/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitThresholds1DayGet(
   DSL_Context_t *pContext,
   DSL_PM_LineInitThreshold_t *pThreshs)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineInitData_t *pThresholds;

   DSL_CHECK_POINTER(pContext, pThreshs);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitThresholds1DayGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pThresholds = DSL_DRV_PM_PTR_LINE_INIT_THRESHOLD_1DAY();

   if( pThresholds )
   {
      memcpy( &(pThreshs->data), pThresholds, sizeof(DSL_PM_LineInitData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - Line Init 1-day thresholds get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERROR;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   if ( (nErrCode == DSL_SUCCESS) && (DSL_DRV_PM_CONTEXT(pContext)->bPmDataValid != DSL_TRUE))
   {
      nErrCode = DSL_WRN_INCOMPLETE_RETURN_VALUES;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitThresholds1DayGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_OPTIONAL_PARAMETERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_THRESHOLDS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_HISTORY*/

#ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitCountersTotalGet(
   DSL_Context_t *pContext,
   DSL_PM_LineInitCountersTotal_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_PM_LineInitData_t *pLinitCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitCountersTotalGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   /* Fill Total Counters elapsed time*/
   pCounters->total.nElapsedTime = DSL_DRV_PM_CONTEXT(pContext)->nPmTotalElapsedTime/DSL_PM_MSEC;

   /* Lock PM module direction mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   pLinitCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_TOTAL();

   if (pLinitCounters != DSL_NULL)
   {
      /* Copy Total Channel Counters data*/
      memcpy(&(pCounters->data), pLinitCounters, sizeof(DSL_PM_LineInitData_t));
   }
   else
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - total counters data pointer is NULL"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      nErrCode = DSL_ERR_INTERNAL;
   }

   /* Unlock PM module direction mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitCountersTotalGet, retCode=%d" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_TOTAL_COUNTERS*/

#ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS
/*
   For a detailed description of the function, its arguments and return value
   please refer to the description in the header file 'drv_dsl_cpe_intern_pm.h'
*/
DSL_uint32_t DSL_DRV_PM_LineInitCountersShowtimeGet(
   DSL_IN DSL_Context_t *pContext,
   DSL_IN_OUT DSL_PM_LineInitCounters_t *pCounters)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_int_t histIdx = -1;
   DSL_uint32_t histFillLevel = 0, histInv = 0;
   DSL_PM_LineInitData_t *pLineCounters;

   DSL_CHECK_POINTER(pContext, pCounters);
   DSL_CHECK_ERR_CODE();

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: IN - DSL_DRV_PM_LineInitCounters15MinGet" DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext)));

   /* Check if the PM module is ready*/
   if (!DSL_DRV_PM_IsPmReady(pContext))
   {
      return DSL_ERROR;
   }

   if( DSL_DRV_PM_CONTEXT(pContext)->bShowtimeProcessingStart == DSL_FALSE )
   {
      DSL_DEBUG(DSL_DBG_WRN,
         (pContext, "DSL[%02d]: No showtime data for the specified interval"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));

      memset( &(pCounters->data), 0x0, sizeof(DSL_PM_LineInitData_t));
      pCounters->interval.bValid       = DSL_FALSE;
      pCounters->interval.nElapsedTime = 0;
      pCounters->interval.nNumber      = 0;

      return DSL_WRN_PM_NO_SHOWTIME_DATA;
   }

   /* Clear the output structure*/
   pCounters->interval.nElapsedTime = 0;
   memset(&(pCounters->data), 0x0, sizeof(DSL_PM_LineInitData_t));

   /* Get history fill level*/
   nErrCode = DSL_DRV_PM_HistoryFillLevelGet(
                 pContext, DSL_DRV_PM_PTR_LINE_HISTORY_SHOWTIME(), &histFillLevel);

   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - history fill level get failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   if( pCounters->nHistoryInterval > histFillLevel )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - no history data for the specified interval(%d)!"
         DSL_DRV_CRLF, DSL_DEV_NUM(pContext), pCounters->nHistoryInterval));
      return DSL_ERROR;
   }

   /* Lock PM module direcetion mutex*/
   nErrCode = DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_TRUE);
   if( nErrCode != DSL_SUCCESS )
   {
      DSL_DEBUG( DSL_DBG_ERR,
         (pContext, "DSL[%02d]: ERROR - PM direction mutex lock failed!"DSL_DRV_CRLF,
         DSL_DEV_NUM(pContext)));
         
      return nErrCode;
   }

   for( histInv = 0; histInv <= pCounters->nHistoryInterval; histInv++ )
   {
      /* Get history item index*/
      nErrCode = DSL_DRV_PM_HistoryItemIdxGet(
                    pContext, DSL_DRV_PM_PTR_LINE_HISTORY_SHOWTIME(),
                    histInv, &histIdx);
      if( nErrCode != DSL_SUCCESS || histIdx < 0 )
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history index get failed!" DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         break;
      }

      /* Set Showtime Elapsed time*/
      pCounters->interval.nElapsedTime +=
         DSL_DRV_PM_CONTEXT(pContext)->pCounters->lineCounters.nShowtimeTimeHist[histIdx];

      pLineCounters = DSL_DRV_PM_PTR_LINE_INIT_COUNTERS_SHOWTIME(histIdx);

      if( pLineCounters != DSL_NULL )
      {
         pCounters->data.nFullInits        += pLineCounters->nFullInits;
         pCounters->data.nFailedFullInits  += pLineCounters->nFailedFullInits;
         pCounters->data.nShortInits       += pLineCounters->nShortInits;
         pCounters->data.nFailedShortInits += pLineCounters->nFailedShortInits;
      }
      else
      {
         DSL_DEBUG( DSL_DBG_ERR,
            (pContext, "DSL[%02d]: ERROR - history interval data pointer is NULL"DSL_DRV_CRLF,
            DSL_DEV_NUM(pContext)));
            
         nErrCode = DSL_ERR_INTERNAL;
         
         break;
      }
   }

   /* Unlock PM module dircetion mutex*/
   DSL_DRV_PM_DirectionMutexControl(pContext, DSL_NEAR_END, DSL_FALSE);

   if( nErrCode == DSL_SUCCESS )
   {
      pCounters->interval.bValid = pCounters->nHistoryInterval == 0 ? DSL_FALSE : DSL_TRUE;
      pCounters->interval.nNumber = pCounters->nHistoryInterval;
   }

   DSL_DEBUG( DSL_DBG_MSG,
      (pContext, "DSL[%02d]: OUT - DSL_DRV_PM_LineInitCountersShowtimeGet, retCode=%d"DSL_DRV_CRLF,
      DSL_DEV_NUM(pContext), nErrCode));

   return nErrCode;
}
#endif /* #ifdef INCLUDE_DSL_CPE_PM_SHOWTIME_COUNTERS*/
#endif /* #ifdef INCLUDE_DSL_CPE_PM_LINE_COUNTERS*/

/** @} DRV_DSL_CPE_PM */

#endif /* INCLUDE_DSL_PM */
