/******************************************************************************

                               Copyright (c) 2007
                            Infineon Technologies AG
                     Am Campeon 1-12; 81726 Munich, Germany

  For licensing information, see the file 'LICENSE' in the root folder of
  this software module.

******************************************************************************/

/*
Includes
*/
#include <signal.h>
#include <stdarg.h>             /* vaargs */

#include "dsl_cpe_control.h"
#include "dsl_cpe_cli.h"
#include "dsl_cpe_cli_console.h"
#include "dsl_cpe_debug.h"
#include "drv_dsl_cpe_api_ioctl.h"
#include "dsl_cpe_simulator.h"

#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
#include "drv_dsl_cpe_cmv_danube.h"
#endif

#define DSL_CPE_STATIC

#include "dsl_cpe_init_cfg.h"

#undef DSL_CCA_DBG_BLOCK
#define DSL_CCA_DBG_BLOCK DSL_CCA_DBG_APP

#ifdef LINUX
struct termios stored_stdout_settings;
struct termios stored_stdin_settings;
#endif /* LINUX */

#ifdef DSL_DEBUG_TOOL_INTERFACE
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>
#include <net/if.h>
#endif

DSL_boolean_t bScriptError = DSL_FALSE;

/* the console only works in forground mode, so no daemonize ! */
#ifndef INCLUDE_DSL_API_CONSOLE

/* 'daemon()' function is not working correctly therefore this feature has been
   disabled here.
   You might use the '&' operator to start the dsl_cpe_control application
   within background */
#undef USE_DAEMONIZE
#if 0
#ifdef __UCLIBC__
#define UCLIBC_VER ((__UCLIBC_MAJOR__*10000) + (__UCLIBC_MINOR__*100) + __UCLIBC_SUBLEVEL__)
#if (UCLIBC_VER >= 926)
      /* Older versions of uClibc than our tested version (0.9.26) seams to have
         problems with the daemon function */
#define USE_DAEMONIZE   1
#endif
#else
   /* assume that other libraries will work! */
/*#define USE_DAEMONIZE   1*/
#endif /* __UCLIBC__ */
#endif

#endif /* INCLUDE_DSL_API_CONSOLE */

#ifdef INCLUDE_DSL_CPE_API_VINAX
DSL_CPE_STATIC  DSL_char_t *sLowLevCfgName = DSL_NULL;
#endif /* #ifdef INCLUDE_DSL_CPE_API_VINAX*/

DSL_CPE_STATIC  const DSL_CPE_EVT_CodeString_t eventString[] =
{
   {DSL_EVENT_I_LINE_FAILURES, "DSL_EVENT_I_LINE_FAILURES"},
   {DSL_EVENT_I_DATA_PATH_FAILURES, "DSL_EVENT_I_DATA_PATH_FAILURES"},
   {DSL_EVENT_I_LINE_THRESHOLD_CROSSING, "DSL_EVENT_I_LINE_THRESHOLD_CROSSING"},
   {DSL_EVENT_I_CHANNEL_THRESHOLD_CROSSING, "DSL_EVENT_I_CHANNEL_THRESHOLD_CROSSING"},
   {DSL_EVENT_I_DATA_PATH_THRESHOLD_CROSSING, "DSL_EVENT_I_DATA_PATH_THRESHOLD_CROSSING"},
   {DSL_EVENT_I_CHANNEL_DATARATE_SHIFT_THRESHOLD_CROSSING, "DSL_EVENT_I_CHANNEL_DATARATE_SHIFT_THRESHOLD_CROSSING"},
   {DSL_EVENT_S_LINIT_FAILURE, "DSL_EVENT_S_LINIT_FAILURE"},
   {DSL_EVENT_S_LINE_STATE, "DSL_EVENT_S_LINE_STATE"},
   {DSL_EVENT_S_LINE_POWERMANAGEMENT_STATE, "DSL_EVENT_S_LINE_POWERMANAGEMENT_STATE"},
   {DSL_EVENT_S_CHANNEL_DATARATE, "DSL_EVENT_S_CHANNEL_DATARATE"},
   {DSL_EVENT_S_FIRMWARE_ERROR, "DSL_EVENT_S_FIRMWARE_ERROR"},
   {DSL_EVENT_S_INIT_READY, "DSL_EVENT_S_INIT_READY"},
   {DSL_EVENT_S_FE_INVENTORY_AVAILABLE, "DSL_EVENT_S_FE_INVENTORY_AVAILABLE"},
   {DSL_EVENT_S_SYSTEM_STATUS, "DSL_EVENT_S_SYSTEM_STATUS"},
   {DSL_EVENT_S_PM_SYNC, "DSL_EVENT_S_PM_SYNC"},
   {DSL_EVENT_S_LINE_TRANSMISSION_STATUS, "DSL_EVENT_S_LINE_TRANSMISSION_STATUS"},
   {DSL_EVENT_S_SHOWTIME_LOGGING, "DSL_EVENT_S_SHOWTIME_LOGGING"},
   {DSL_EVENT_S_FIRMWARE_REQUEST, "DSL_EVENT_S_FIRMWARE_REQUEST"},
   {DSL_EVENT_S_FIRMWARE_DOWNLOAD_STATUS, "DSL_EVENT_S_FIRMWARE_DOWNLOAD_STATUS"},
   {DSL_EVENT_S_AUTOBOOT_STATUS, "DSL_EVENT_S_AUTOBOOT_STATUS"},
   {DSL_EVENT_S_SNMP_MESSAGE_AVAILABLE, "DSL_EVENT_S_SNMP_MESSAGE_AVAILABLE"},
   {DSL_EVENT_S_SYSTEM_INTERFACE_STATUS, "DSL_EVENT_S_SYSTEM_INTERFACE_STATUS"},
   {DSL_EVENT_LAST, DSL_NULL}
};

/*
local prototypes
*/
DSL_CPE_STATIC  DSL_void_t DSL_CPE_ArgParse (
   DSL_int32_t argc,
   DSL_char_t * argv[]
);

DSL_CPE_STATIC  DSL_void_t DSL_CPE_Help (
   DSL_char_t * sApplicationName
);

DSL_CPE_STATIC  DSL_void_t DSL_CPE_Termination (
);


DSL_CPE_STATIC  DSL_CPE_Control_Context_t *gDSLContext = DSL_NULL;

#ifndef DSL_CPE_REMOVE_PIPE_SUPPORT
extern DSL_Error_t DSL_CPE_Pipe_Init (DSL_CPE_Control_Context_t * pContext);
#endif

DSL_char_t *g_sFirmwareName1 = DSL_NULL;
DSL_char_t *g_sFirmwareName2 = DSL_NULL;
#ifdef INCLUDE_SCRIPT_NOTIFICATION
DSL_char_t *g_sRcScript = DSL_NULL;
DSL_CPE_STATIC  DSL_boolean_t bScriptWarn = DSL_FALSE;
#endif

DSL_boolean_t g_bWaitBeforeLinkActivation = DSL_FALSE;
DSL_boolean_t g_bWaitBeforeConfigWrite    = DSL_FALSE;
DSL_boolean_t g_bWaitBeforeRestart        = DSL_FALSE;

static DSL_LineStateValue_t g_nPrevLineState = DSL_LINESTATE_NOT_INITIALIZED;

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
DSL_char_t *g_sAdslScript = DSL_NULL;
DSL_char_t *g_sVdslScript = DSL_NULL;
DSL_boolean_t g_bAutoContinueWaitBeforeLinkActivation = DSL_TRUE;
DSL_boolean_t g_bAutoContinueWaitBeforeConfigWrite    = DSL_TRUE;
DSL_boolean_t g_bAutoContinueWaitBeforeRestart        = DSL_TRUE;
#endif
#endif

/*
local variables
*/
DSL_CPE_STATIC  DSL_int32_t bHelp = -1;
DSL_CPE_STATIC  DSL_int32_t bGetVersion = -1;
DSL_CPE_STATIC  DSL_int32_t bInit = -1;
DSL_CPE_STATIC  DSL_int32_t bXtuOctets = -1;

/* Events and resource handling is enabled for all available types by default! */
DSL_CPE_STATIC  DSL_boolean_t bEventActivation = DSL_TRUE;
DSL_CPE_STATIC  DSL_uint32_t nResourceActivationMask = 0x00000000;

#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   DSL_int32_t bOptimize = 1;
#else
   DSL_int32_t bOptimize = -1;
#endif
DSL_CPE_STATIC  DSL_int32_t bMsgDump = -1;

#ifdef DSL_DEBUG_TOOL_INTERFACE
DSL_CPE_STATIC  DSL_int32_t bTcpMessageIntf = -1;
#endif

#ifdef USE_DAEMONIZE
DSL_CPE_STATIC  DSL_int_t bNotSilent = 1;
#endif /* USE_DAEMONIZE */

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
DSL_CPE_STATIC  DSL_int32_t bConsole = -1;
#endif

DSL_CPE_STATIC  DSL_int_t g_bFirmware1 = -1;
DSL_CPE_STATIC  DSL_int_t g_bFirmware2 = -1;

DSL_CPE_STATIC  DSL_G997_XTUSystemEnablingData_t g_nXtseInit;
DSL_CPE_STATIC  DSL_uint8_t g_nMsgDumpDbgLvl;

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
typedef struct
{
   /** reference to registered CLI */
   DSL_CLI_Context_t *pCLIContext;
   /** running status of pipe task */
   volatile DSL_boolean_t bRun;
} dummy_console_t;

DSL_CPE_STATIC  dummy_console_t dummy_console;

DSL_CPE_STATIC  DSL_Error_t DSL_CPE_Control_Exit (DSL_void_t * pContext);
#endif

#ifdef LINUX
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "/firmware/modemhwe.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  ""
#elif defined(INCLUDE_DSL_CPE_API_VINAX)
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "/opt/ifx/firmware/vcpe_hw.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  "/opt/ifx/firmware/acpe_hw.bin"
#else
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "firmware0.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  "firmware1.bin"
#endif
#elif VXWORKS
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "modemhwe.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  ""
#elif defined(INCLUDE_DSL_CPE_API_VINAX)
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "vcpe_hw.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  "acpe_hw.bin"
#else
   #define DSL_CPE_DEFAULT_FIRMWARE_1  "firmware0.bin"
   #define DSL_CPE_DEFAULT_FIRMWARE_2  "firmware1.bin"
#endif
#else
   #define DSL_CPE_DEFAULT_FIRMWARE_1  ""
   #define DSL_CPE_DEFAULT_FIRMWARE_2  ""
#endif

const DSL_char_t *sDefaultFirmwareName1 = (DSL_char_t *)DSL_CPE_DEFAULT_FIRMWARE_1;
const DSL_char_t *sDefaultFirmwareName2 = (DSL_char_t *)DSL_CPE_DEFAULT_FIRMWARE_2;

#ifdef INCLUDE_SCRIPT_NOTIFICATION
   #if defined (INCLUDE_DSL_CPE_API_DANUBE)
      #define DSL_CPE_DEFAULT_RC_SCRIPT "./adslrc.sh"
   #elif defined(INCLUDE_DSL_CPE_API_VINAX)
      #define DSL_CPE_DEFAULT_RC_SCRIPT "./vdslrc.sh"
   #else
      #define DSL_CPE_DEFAULT_RC_SCRIPT "./xdslrc.sh"
   #endif
   const DSL_char_t *sDefaultRcScript = DSL_CPE_DEFAULT_RC_SCRIPT;
#endif

#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
DSL_char_t *sSoapRemoteServer = DSL_NULL;
#endif

#ifdef DSL_DEBUG_TOOL_INTERFACE
DSL_char_t *sTcpMessagesSocketAddr = DSL_NULL;
#endif

DSL_InitData_t gInitCfgData;


DSL_CPE_STATIC  struct option long_options[] = {
   {"help      ", 0, 0, 'h'},
   {"version   ", 0, 0, 'v'},
   {"init      ", 0, 0, 'i'},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"device_num", 1, 0, 'd'},
   {"low_cfg   ", 1, 0, 'l'},
#endif
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   {"console   ", 0, 0, 'c'},
#endif
   {"event_cnf ", 1, 0, 'e'},
   {"msg_dump  ", 0, 0, 'm'},
#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   {"auto_scr_1", 0, 0, 'a'},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"auto_scr_2", 0, 0, 'A'},
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#endif
#endif
#ifdef USE_DAEMONIZE
   {"silent    ", 0, 0, 'q'},
#endif /* USE_DAEMONIZE */
   {"firmware1 ", 1, 0, 'f'},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"firmware2 ", 1, 0, 'F'},
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   {"opt_off   ", 1, 0, 'o'},
#endif
#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
   {"soap      ", 0, 0, 's'},
#endif
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   {"notif     ", 1, 0, 'n'},
#endif
#ifdef DSL_DEBUG_TOOL_INTERFACE
   {"tcpmsg    ", 2, 0, 't'},
#endif
   {0, 0, 0, 0}
};

/* 1 colon means there is a required parameter */
/* 2 colons means there is an optional parameter */
DSL_CPE_STATIC  const DSL_char_t GETOPT_LONG_OPTSTRING[] = "hvi::"
#ifdef INCLUDE_DSL_CPE_API_VINAX
   "d:"
   "l:"
#endif
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   "c"
#endif
   "e:m::"
#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   "a:"
#ifdef INCLUDE_DSL_CPE_API_VINAX
   "A:"
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#endif
#endif
#ifdef USE_DAEMONIZE
   "q"
#endif
   "f:"
#ifdef INCLUDE_DSL_CPE_API_VINAX
   "F:"
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   "o"
#endif
#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
   "s:"
#endif
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   "n::"
#endif
#ifdef DSL_DEBUG_TOOL_INTERFACE
   "t::"
#endif
   ;

/*lint -save -e786 */ \
DSL_CPE_STATIC  DSL_char_t description[][90] = {
   {"help screen"},
   {"display version"},
   {"init device w/ <xtu> Bits seperated by underscore (e.g. -i05_01_04_00_04_01_00_00)"},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"device number (e.g. -d0)"},
   {"low level configuration file"},
#endif
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   {"start console"},
#endif
   {"configure instance activation handling <enable/disable>[_mask] (e.g. -e1_1)"},
   {"enable message dump"},
#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   {"autoboot start script for ADSL (empty by default)"},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"autoboot start script for VDSL (empty by default)"},
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#endif
#endif
#ifdef USE_DAEMONIZE
   "silent mode, no output from background",
#endif
   {"firmware file, default " DSL_CPE_DEFAULT_FIRMWARE_1},
#ifdef INCLUDE_DSL_CPE_API_VINAX
   {"2nd firmware file, default " DSL_CPE_DEFAULT_FIRMWARE_2},
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   {"deactivate footprint optimizations"},
#endif
#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
   {"SOAP server"},
#endif
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   {"notification script name, default " DSL_CPE_DEFAULT_RC_SCRIPT},
#endif
#ifdef DSL_DEBUG_TOOL_INTERFACE
   {"enable dbgtool support on tcp ports 2000/2001, listen only on <ipaddr> (optional)"},
#endif
   {0}
};

/*lint -restore */

#ifndef DSL_CPE_DEBUG_DISABLE
DSL_CPE_STATIC  const DSL_char_t *dsl_cpe_ctl_version = "@(#)" PACKAGE_VERSION;
#endif

DSL_CPE_Control_Context_t *DSL_CPE_GetGlobalContext (
   DSL_void_t
)
{
   if (gDSLContext == DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "ERROR, requesting for global context, which is zero" DSL_CPE_CRLF));
   }

   return gDSLContext;
}

void DSL_CPE_Echo (
   DSL_char_t *buf)
{
   DSL_char_t *str;
   DSL_char_t cmd[64];

   /* remove leading whitespaces */
   for (str = buf; str && *str && isspace((int)(*str)); ++str)
   {
      ;
   }

   sscanf (str, "%s", &cmd);
   if (strcmp (cmd, "echo") == 0)
   {
      str += 4;
   }
   if (str[0] == ' ') str++;

   DSL_CPE_FPrintf (DSL_CPE_STDOUT, "%s", str);

   return;
}

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
/* Use this to format arrays of unspecified contents as hex values */
DSL_void_t DSL_CPE_ArraySPrintF(
   DSL_char_t *pDst,
   DSL_void_t *pSrc,
   DSL_uint16_t nSrcSize,
   DSL_uint16_t nSrcElementSize,
   DSL_CPE_ArrayPrintFormat_t nFormat)
{
   DSL_uint16_t i = 0, j = 0, elements = 0;
   DSL_uint32_t nVal = 0;
   DSL_char_t c;
   DSL_int_t ret = 0;

   elements = nSrcSize/nSrcElementSize;

   if (nFormat == DSL_ARRAY_FORMAT_HEX)
   {
      ret = sprintf(pDst, "(");
      pDst++;
   }

   for (i = 0; i < elements; i++)
   {
      if ((i != 0) && (nFormat == DSL_ARRAY_FORMAT_HEX))
      {
         ret = sprintf(pDst, ",");
         pDst ++;
      }

      if (nFormat == DSL_ARRAY_FORMAT_HEX)
      {
         switch (nSrcElementSize)
         {
         case 1:
            ret = sprintf(pDst, "%02X", ((DSL_uint8_t*)pSrc)[i]);
            break;
         case 2:
            ret = sprintf(pDst, "%04X", ((DSL_uint16_t*)pSrc)[i]);
            break;
         case 4:
            ret = sprintf(pDst, "%08X", ((DSL_uint32_t*)pSrc)[i]);
            break;
         default:
            ret = sprintf(pDst, "xx");
            break;
         }

         if(ret > 0)
         {
            pDst += ret;
         }
      }
      else
      {
         switch (nSrcElementSize)
         {
         case 1:
            nVal = (DSL_uint32_t)(((DSL_uint8_t*)pSrc)[i]);
            break;
         case 2:
            nVal = (DSL_uint32_t)(((DSL_uint16_t*)pSrc)[i]);
            break;
         case 4:
            nVal = (DSL_uint32_t)(((DSL_uint32_t*)pSrc)[i]);
            break;
         default:
            nVal = 0;
            break;
         }

         for (j = 0; j < nSrcElementSize; j++)
         {
            c = (DSL_char_t)(nVal >> j);
            if (isprint(c) != 0)
            {
               ret = sprintf(pDst, "%c", c);
               pDst += ret;
            }
         }
      }
   }

   if (nFormat == DSL_ARRAY_FORMAT_HEX)
   {
      ret = sprintf(pDst, ")");
   }
}

/**
   Parses the given character string for given token and sets the resulting
   pointer to the given offset of tokens within the given string.

   \param DSL_char_t *pCommands
      Specifies a string that shall be parsed
   \param DSL_int_t nParamNr
      Specifies the number of tokens that shall be found and of which the
      pointer has to be incresed.
   \param DSL_char_t *pSeps
      Specifies the separators to be srearched for
   \param DSL_char_t *pCmdOffset
      Returns a pointer to the command string including token offset if found
*/
DSL_Error_t DSL_CPE_MoveCharPtr(
   DSL_char_t *pCommands,
   DSL_int_t nParamNr,
   DSL_char_t *pSeps,
   DSL_char_t **pCmdOffset)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_int_t i = 0;
   DSL_char_t *pTmp = DSL_NULL;

   pTmp = pCommands;

   for (i = 0; i < nParamNr; i++)
   {
      if (pTmp == DSL_NULL) break;

      pTmp = strpbrk(pTmp, pSeps);
      if (pTmp != DSL_NULL)
      {
         pTmp++;
      }
      else
      {
         break;
      }
   }

   if (i < nParamNr)
   {
      nRet = DSL_ERROR;
      *pCmdOffset = pCommands;
   }
   else
   {
      nRet = DSL_SUCCESS;
      *pCmdOffset = pTmp;
   }

   return nRet;
}


#ifdef INCLUDE_DSL_CPE_API_VINAX
DSL_Error_t DSL_CPE_GetMacAdrFromString(
   DSL_char_t *pString,
   DSL_CPE_MacAddress_t *pMacAdr)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_char_t string[20] = { 0 };
   DSL_char_t seps[] = ":";
   DSL_char_t *token;
   DSL_int_t i = 0;

   strncpy (string, pString, sizeof(string)-1);
   string[sizeof(string)-1] = 0;

   /* Get first token */
   token = strtok (string, seps);
   if (token != DSL_NULL)
   {
      for (i = 0; i < DSL_MAC_ADDRESS_OCTETS; i++)
      {
         DSL_CPE_sscanf (token, "%bx", (unsigned char *)&(pMacAdr->nAdr[i]));

         /* Get next token */
         token = strtok(DSL_NULL, seps);

         /* Exit scanning if no further information is included */
         if (token == DSL_NULL)
         {
            break;
         }
      }
   }

   if (i < (DSL_MAC_ADDRESS_OCTETS - 1))
   {
      nRet = DSL_ERROR;
   }
   else
   {
      nRet = DSL_SUCCESS;
   }

   return nRet;
}
#endif /* INCLUDE_DSL_CPE_API_VINAX */
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

/**
   Parse all arguments and enable requested features.

   \param argc - number of parameters
   \param argv - array of parameter strings
*/
DSL_CPE_STATIC  DSL_void_t DSL_CPE_ArgParse (
   DSL_int32_t argc,
   DSL_char_t * argv[]
)
{
   DSL_int32_t parm_no = 0;
   DSL_char_t *pEndPtr = "\0";

   DSL_char_t string[30] = { 0 };
   DSL_int_t i = 0;
   DSL_uint32_t nVal = 0;
   DSL_char_t seps[]   = "_";
   DSL_char_t *token, *errMask;
   DSL_int_t option_index = 1;

   while (1)
   {
      DSL_int_t c;

      /* 1 colon means there is a required parameter */
      /* 2 colons means there is an optional parameter */
      c = getopt_long (argc, argv, GETOPT_LONG_OPTSTRING, long_options,
         &option_index);

      if (c == -1)
      {
         break;
      }
      
      parm_no++;

      switch (c)
      {
      case 0:
      /* optarg found */
      break;
      
      case 'h':
         bHelp = 1;
         break;

      case 'v':
         bGetVersion = 1;
         break;

#ifdef DSL_DEBUG_TOOL_INTERFACE
      case 't':
         bTcpMessageIntf = 1;
         if (optarg != NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
               "using %s as an IP address for tcp messages debug server" DSL_CPE_CRLF,
               optarg));

            sTcpMessagesSocketAddr = DSL_CPE_Malloc(strlen (optarg) + 1);
            if (sTcpMessagesSocketAddr)
            {
               strcpy (sTcpMessagesSocketAddr, optarg);
            }
         }
         else
         {
            sTcpMessagesSocketAddr = DSL_CPE_OwnAddrStringGet();
            if (sTcpMessagesSocketAddr == DSL_NULL)
            {
               DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                  "Own IP address get faiuled!" DSL_CPE_CRLF));
            }
            else
            {
               DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                  "using %s as an IP address for tcp messages debug server" DSL_CPE_CRLF,
                  sTcpMessagesSocketAddr));
            }
         }
         break;
#endif

#ifdef USE_DAEMONIZE
      case 'q':
         bNotSilent = 0;
         break;
#endif /* USE_DAEMONIZE */

      case 'i':
         bInit = 1;
         memset (&g_nXtseInit.XTSE, 0, sizeof(g_nXtseInit.XTSE));

         if (optarg != DSL_NULL)
         {
            strncpy (string, optarg, sizeof(string)-1);
            string[sizeof(string)-1]=0;

            token = strtok (string, seps);
            if (token != DSL_NULL)
            {
               for (i = 0; i < DSL_G997_NUM_XTSE_OCTETS; i++)
               {
                  DSL_CPE_sscanf (token, "%x", &nVal);
                  g_nXtseInit.XTSE[i] = (DSL_uint8_t) nVal;

                  /* Get next token */
                  token = strtok(DSL_NULL, seps);

                  /* Exit scanning if no further information is included */
                  if (token == DSL_NULL)
                  {
                     break;
                  }
               }
            }
            if (i > 0) bXtuOctets = 1;
         }
         break;

#ifdef INCLUDE_DSL_CPE_API_VINAX      
     case 'd':
         if (optarg != DSL_NULL)
         {
            DSL_CPE_sscanf (optarg, "%u", &(DSL_CPE_GetGlobalContext()->nDevNum));
         }
         break;

     case 'l':
         if (sLowLevCfgName)
         {
            DSL_CPE_Free(sLowLevCfgName);
         }
         sLowLevCfgName = DSL_CPE_Malloc (strlen (optarg) + 1);
         strcpy (sLowLevCfgName, optarg);
         break;
#endif

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
      case 'c':
         bConsole = 1;
         break;
#endif

#ifdef INCLUDE_SCRIPT_NOTIFICATION
      case 'n':
         if (g_sRcScript)
         {
            DSL_CPE_Free(g_sRcScript);
         }

         if (optarg != NULL)
         {
            g_sRcScript = DSL_CPE_Malloc (strlen (optarg) + 1);
            if (g_sRcScript)
            {
               strcpy (g_sRcScript, optarg);
            }
         }
         else
         {
            g_sRcScript = DSL_CPE_Malloc (strlen (sDefaultRcScript) + 1);
            if (g_sRcScript)
            {
               strcpy (g_sRcScript, sDefaultRcScript);
            }
         }
         break;
#endif /* #ifdef INCLUDE_SCRIPT_NOTIFICATION */

      case 'f':
         g_bFirmware1 = 1;
         if (g_sFirmwareName1)
         {
            DSL_CPE_Free(g_sFirmwareName1);
         }
         g_sFirmwareName1 = DSL_CPE_Malloc (strlen (optarg) + 1);
         if (g_sFirmwareName1)
         {
            strcpy (g_sFirmwareName1, optarg);
         }
         break;

      case 'F':
         g_bFirmware2 = 1;
         if (g_sFirmwareName2)
         {
            DSL_CPE_Free(g_sFirmwareName2);
         }
         g_sFirmwareName2 = DSL_CPE_Malloc (strlen (optarg) + 1);
         if (g_sFirmwareName2)
         {
            strcpy (g_sFirmwareName2, optarg);
         }
         break;

#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_SE)
      case 'o':
         bOptimize = -1;
         break;
#endif

      case 'e':
         if (optarg != NULL)
         {
            strncpy (string, optarg, sizeof(string)-1);
            string[sizeof(string)-1]=0;
            token = strtok (string, seps);

            if (token != DSL_NULL)
            {
               /* Get events enable/disable flag*/
               DSL_CPE_sscanf (token, "%d", &nVal);
               bEventActivation = nVal ? DSL_TRUE : DSL_FALSE;

               /* Get next token, event mask */
               token = strtok(DSL_NULL, seps);
               if (token != DSL_NULL)
               {
                  nResourceActivationMask = strtoul (token, &errMask, 16);

                  if (nResourceActivationMask == ULONG_MAX && errno != 0)
                  {
                     DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                        "invalid resource mask %s specified, all resources will "
                        "be disabled" DSL_CPE_CRLF, optarg));
                     nResourceActivationMask = 0xFFFFFFFF;
                  }
               }
            }

            if (bEventActivation == DSL_FALSE)
            {
               /* Mask (disable) all resources */
               nResourceActivationMask = 0xFFFFFFFF;               
            }
         }
         break;

      case 'm':
         bMsgDump = 1;
         if (optarg != DSL_NULL)
         {
            g_nMsgDumpDbgLvl = (DSL_uint8_t)(strtoul (optarg, &pEndPtr, 0));
         }
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "Using message dump (debug level: 0x%02X)" DSL_CPE_CRLF, g_nMsgDumpDbgLvl));
         break;

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
      case 'a':
         if (g_sAdslScript)
         {
            DSL_CPE_Free(g_sAdslScript);
            g_sAdslScript = DSL_NULL;
         }
         g_sAdslScript = DSL_CPE_Malloc (strlen (optarg) + 1);
         if (g_sAdslScript)
         {
            strcpy (g_sAdslScript, optarg);
         }
         break;

      case 'A':

         if (g_sVdslScript)
         {
            DSL_CPE_Free(g_sVdslScript);
            g_sVdslScript = DSL_NULL;
         }
         g_sVdslScript = DSL_CPE_Malloc (strlen (optarg) + 1);
         if (g_sVdslScript)
         {
            strcpy (g_sVdslScript, optarg);
         }
         break;
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */

#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
      case 's':
         if (optarg != NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
               "using SOAP server - %s" DSL_CPE_CRLF, optarg));

            sSoapRemoteServer = DSL_CPE_Malloc(strlen (optarg) + 1);
            if (sSoapRemoteServer)
            {
               strcpy (sSoapRemoteServer, optarg);
            }
         }
         break;
#endif
      default:
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "Sorry, there is a unrecognized option: %c" DSL_CPE_CRLF, c));
         break;
      }
   }

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   if( (g_sAdslScript != DSL_NULL)
#ifdef INCLUDE_DSL_CPE_API_VINAX
      || (g_sVdslScript != DSL_NULL)
#endif
      )
   {
      g_bWaitBeforeLinkActivation = DSL_TRUE;
      g_bWaitBeforeConfigWrite    = DSL_TRUE;
      g_bWaitBeforeRestart        = DSL_TRUE;
   }
   else
   {
      g_bWaitBeforeLinkActivation = DSL_FALSE;
      g_bWaitBeforeConfigWrite    = DSL_FALSE;
      g_bWaitBeforeRestart        = DSL_FALSE;
   }
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */
}

/**
Print usage help.

\param sApplicationName    name of the application

*/
DSL_CPE_STATIC  DSL_void_t DSL_CPE_Help (
   DSL_char_t *sApplicationName
)
{
   struct option *ptr;
   DSL_char_t *desc = description[0];

   ptr = long_options;

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
      "Welcome to DSL CPI API control application" DSL_CPE_CRLF));

   if (strlen(sApplicationName) == 0)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "usage: %s [options]" DSL_CPE_CRLF, sApplicationName));
   }
   else
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "usage: [options]" DSL_CPE_CRLF, sApplicationName));
   }
   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
      "following options are available:" DSL_CPE_CRLF));
#ifdef DEBUG
   sleep (1);
#endif

   while (ptr->name)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         " --%s\t(-%c)\t- %s" DSL_CPE_CRLF, ptr->name, ptr->val, desc));
#ifdef DEBUG
      sleep (1);
#endif
      ptr++;
      desc += sizeof (description[0]);
   }
   return;
}

#ifdef DSL_CPE_SOAP_FW_UPDATE
DSL_Error_t DSL_CPE_SoapFirmwareUpdate(
   DSL_CPE_Firmware_t *pFirmware,
   DSL_CPE_Firmware_t *pFirmware2)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_CPE_Control_Context_t *pCtrlCtx = DSL_NULL;

   pCtrlCtx = DSL_CPE_GetGlobalContext();

   if (pCtrlCtx == DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "Invalid context pointer!" DSL_CPE_CRLF));
      return DSL_ERROR;
   }

   DSL_CPE_LockGet(&pCtrlCtx->semFwUpdate);

   nRet = DSL_CPE_SoapFirmwareStore(pCtrlCtx, pFirmware, &(pCtrlCtx->firmware));
   nRet = DSL_CPE_SoapFirmwareStore(pCtrlCtx, pFirmware2, &(pCtrlCtx->firmware2));

   DSL_CPE_LockSet(&pCtrlCtx->semFwUpdate);

   return nRet;
}

DSL_Error_t DSL_CPE_SoapFirmwareStore(
   DSL_CPE_Control_Context_t *pContext,
   DSL_CPE_Firmware_t *pSrcFirmware,
   DSL_CPE_Firmware_t *pDestFirmware)
{

   if ((pSrcFirmware->pData != DSL_NULL) && (pSrcFirmware->nSize > 0))
   {
      if (pDestFirmware->pData != DSL_NULL)
      {
         DSL_CPE_Free(pDestFirmware->pData);
         pDestFirmware->pData = DSL_NULL;
         pDestFirmware->nSize = 0;
      }

      pDestFirmware->pData = DSL_CPE_Malloc(pSrcFirmware->nSize);
      if (pDestFirmware->pData != DSL_NULL)
      {
         memcpy(pDestFirmware->pData, pSrcFirmware->pData, pSrcFirmware->nSize);
         pDestFirmware->nSize = pSrcFirmware->nSize;
      }
      else
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "Failed to allocate memory to store firmware binary of size %d bytes"
            DSL_CPE_CRLF, pSrcFirmware->nSize));
         return DSL_ERROR;
      }
   }

   return DSL_SUCCESS;
}
#endif /* DSL_CPE_SOAP_FW_UPDATE */

DSL_Error_t DSL_CPE_LoadFirmwareFromFile(
   DSL_char_t *psFirmwareName,
   DSL_uint8_t **pFirmware,
   DSL_uint32_t *pnFirmwareSize)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_CPE_File_t *fd_image = DSL_NULL;
   DSL_CPE_stat_t file_stat;

   *pnFirmwareSize = 0;
   if (*pFirmware != DSL_NULL)
   {
      DSL_CPE_Free(*pFirmware);
      *pFirmware = DSL_NULL;
   }

   for (;;)
   {
      fd_image = DSL_CPE_FOpen (psFirmwareName, "rb");
      if (fd_image == DSL_NULL)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
            " open %s fail." DSL_CPE_CRLF, psFirmwareName));
         nRet = DSL_ERROR;
         break;
      }

      DSL_CPE_FStat (psFirmwareName, &file_stat);
      *pnFirmwareSize = file_stat.st_size;
      *pFirmware = DSL_CPE_Malloc (*pnFirmwareSize);
      if (DSL_CPE_FRead (*pFirmware, sizeof (DSL_uint8_t), *pnFirmwareSize, fd_image) <= 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
            "\n firmware_image not present"));
         nRet = DSL_ERROR;
         break;
      }

      break;
   }

   if (fd_image != DSL_NULL)
   {
      DSL_CPE_FClose (fd_image);
      fd_image = DSL_NULL;
   }

   return nRet;
}

DSL_Error_t DSL_CPE_DownloadFirmware(
   DSL_FirmwareRequestType_t nFwReqType,
   DSL_char_t *pcFw,
   DSL_char_t *pcFw2)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_CPE_Control_Context_t *pContext = DSL_NULL;
   DSL_AutobootLoadFirmware_t ldFw;
   DSL_AutobootControl_t pAcs;
   DSL_AutobootStatus_t pAsg;
   DSL_CPE_File_t *fd_image = DSL_NULL;
   DSL_CPE_stat_t file_stat;
   DSL_int_t nSize = 0, nReadSize = DSL_CPE_FW_CHUNK_SIZE;
   DSL_int_t nWriteSize = 0;
   DSL_uint8_t *pChunkData = DSL_NULL;
   DSL_boolean_t bSetFw1 = DSL_TRUE;
#if defined(INCLUDE_DSL_CPE_API_VINAX)   
   DSL_boolean_t bSetFw2 = DSL_TRUE;
#endif   

   memset(&ldFw, 0, sizeof(DSL_AutobootLoadFirmware_t));

   pContext = DSL_CPE_GetGlobalContext();
   if (pContext == DSL_NULL)
   {
      return DSL_ERROR;
   }

#ifdef DSL_CPE_SOAP_FW_UPDATE
   DSL_CPE_LockGet(&pContext->semFwUpdate);

   /*
      If firmware binaries are given by SOAP interface this will be used in any
      case.
   */
   if ((pContext->firmware.pData != DSL_NULL) && (pContext->firmware.nSize > 0))
   {
      ldFw.data.pFirmware = pContext->firmware.pData;
      ldFw.data.nFirmwareSize = pContext->firmware.nSize;
      pcFw = DSL_NULL;
      bSetFw1 = DSL_FALSE;
   }

#if defined(INCLUDE_DSL_CPE_API_VINAX)
   if ((pContext->firmware2.pData != DSL_NULL) && (pContext->firmware2.nSize > 0))
   {
      ldFw.data.pFirmware2 = pContext->firmware2.pData;
      ldFw.data.nFirmwareSize2 = pContext->firmware2.nSize;
      pcFw2 = DSL_NULL;
      bSetFw2 = DSL_FALSE;
   }
#endif /* INCLUDE_DSL_CPE_API_VINAX*/

#endif /* DSL_CPE_SOAP_FW_UPDATE */

   /*
      If firmware name is not defined by current call and no SOAP defined
      firmware is available choose last defined file (or default if startup
      value was not changed).
   */
   if ( (pcFw == DSL_NULL) && (bSetFw1 == DSL_TRUE) )
   {
      if ((g_bFirmware1 != -1) || (strlen(g_sFirmwareName1) > 0))
      {
         pcFw = g_sFirmwareName1;
      }
   }

#if defined(INCLUDE_DSL_CPE_API_VINAX)
   if ( (pcFw2 == DSL_NULL) && (bSetFw2 == DSL_TRUE) )
   {
      if ((g_bFirmware2 != -1) || (strlen(g_sFirmwareName2) > 0))
      {
         pcFw2 = g_sFirmwareName2;
      }
   }
#endif

   do
   {
      if ( (bOptimize == 1) && (bSetFw1 == DSL_TRUE) )
      {
         /*
            Footprint optimized download uses chunks for firmware binary download.
            This is to reduce the necessary runtime memory allocation for download
            procedure
         */

         DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX "Using optimized fw "
            "download..." DSL_CPE_CRLF));

         if (pcFw == DSL_NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
               " firmware file not found!" DSL_CPE_CRLF, pcFw));
            nRet = DSL_ERROR;
            break;
         }

         fd_image = DSL_CPE_FOpen (pcFw, "r");
         if (fd_image == DSL_NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX " open %s failed."
               DSL_CPE_CRLF, pcFw));
            nRet = DSL_ERROR;
            break;
         }

         DSL_CPE_FStat (pcFw, &file_stat);
         nSize = file_stat.st_size;

         pChunkData = DSL_CPE_Malloc (nReadSize);
         if (pChunkData == DSL_NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX "Memory allocation "
               "failed (optimized FW download)" DSL_CPE_CRLF));
            nRet = DSL_ERROR;
            break;
         }
      
         ldFw.data.pFirmware = pChunkData;

         while (nSize > 0)
         {
            if (nSize > DSL_CPE_FW_CHUNK_SIZE)
            {
               nReadSize = DSL_CPE_FW_CHUNK_SIZE;
            }
            else
            {
               nReadSize = nSize;
            }

            if (DSL_CPE_FRead (pChunkData, sizeof (DSL_uint8_t), nReadSize,
               fd_image) <= 0)
            {
               DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                  "firmware_image not present" DSL_CPE_CRLF));
               nRet = DSL_ERROR;
               break;
            }
            
            ldFw.data.nFirmwareSize = nReadSize;
            ldFw.data.nFirmwareOffset = nWriteSize;

            if (nSize - nReadSize <= 0)
            {
               ldFw.data.bLastChunk = DSL_TRUE;
            }
            else
            {
               ldFw.data.bLastChunk = DSL_FALSE;
            }

            nRet = (DSL_Error_t) DSL_CPE_Ioctl(pContext->fd,
               DSL_FIO_AUTOBOOT_LOAD_FIRMWARE, (DSL_int_t) &ldFw);

            if (nRet < DSL_SUCCESS)
            {
               DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                  "Autoboot Load Firmware (using chunks) failed!, nRet = %d!"
                  DSL_CPE_CRLF, ldFw.accessCtl.nReturn));

               break;
            }

            nSize -= nReadSize;
            nWriteSize += nReadSize;
         }

         if (nRet < DSL_SUCCESS)
         {
            break;
         }
         else
         {
            /* $$TD: Temporary only because better solution would be to start
               the link directly within driver after last chunk is downloaded */
            DSL_CPE_Sleep(1);
            nRet = (DSL_Error_t) DSL_CPE_Ioctl(pContext->fd,
               DSL_FIO_AUTOBOOT_STATUS_GET, (DSL_int_t) &pAsg);

            if (pAsg.data.nStatus == DSL_AUTOBOOT_STATUS_STOPPED)
            {
               pAcs.data.nCommand = DSL_AUTOBOOT_CTRL_START;
               nRet = (DSL_Error_t) DSL_CPE_Ioctl(pContext->fd,
                  DSL_FIO_AUTOBOOT_CONTROL_SET, (DSL_int_t) &pAcs);
            }
         }

         /* also store the new firmware binary name within global
            configuration to be used for next download */
         if (pcFw != g_sFirmwareName1)
         {
            /* also store the new firmware binary name within global
            configuration to be used for next download */
            if (ldFw.data.pFirmware != DSL_NULL)
            {
               DSL_CPE_Free(g_sFirmwareName1);
               g_sFirmwareName1 = DSL_CPE_Malloc (strlen (pcFw) + 1);
               if (g_sFirmwareName1)
               {
                  strcpy (g_sFirmwareName1, pcFw);
               }
            }
         }
      }
      else
      {
         /* simple download without any memory optimizations just loads the
            necessary firmware binaries completely into memory. */

         DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX "Using normal fw "
            "download..." DSL_CPE_CRLF));

#if defined(INCLUDE_DSL_CPE_API_VINAX)
         if ( (nFwReqType == DSL_FW_REQUEST_VDSL) || (nFwReqType == DSL_FW_REQUEST_NA) )
#else
         if ( (nFwReqType == DSL_FW_REQUEST_ADSL) || (nFwReqType == DSL_FW_REQUEST_NA) )
#endif
         {
            if (pcFw != DSL_NULL)
            {
               nRet = DSL_CPE_LoadFirmwareFromFile(pcFw, &ldFw.data.pFirmware,
                  &ldFw.data.nFirmwareSize);
               if (nRet < DSL_SUCCESS)
               {
                  nRet = DSL_ERROR;
               }
               else
               {
                  if (pcFw != g_sFirmwareName1)
                  {
                     /* also store the new firmware binary name within global
                        configuration to be used for next download */
                     if (ldFw.data.pFirmware != DSL_NULL)
                     {
                        DSL_CPE_Free(g_sFirmwareName1);
                        g_sFirmwareName1 = DSL_CPE_Malloc (strlen (pcFw) + 1);
                        if (g_sFirmwareName1)
                        {
                           strcpy (g_sFirmwareName1, pcFw);
                        }
                     }
                  }
               }
            }
         }
#if defined(INCLUDE_DSL_CPE_API_VINAX)
         if ( (nFwReqType == DSL_FW_REQUEST_ADSL)  || (nFwReqType == DSL_FW_REQUEST_NA) )
         {
            if (pcFw2 != DSL_NULL)
            {
               nRet = DSL_CPE_LoadFirmwareFromFile(pcFw2, &ldFw.data.pFirmware2,
                  &ldFw.data.nFirmwareSize2);
               if (nRet < DSL_SUCCESS)
               {
                  nRet = DSL_ERROR;
               }
               else
               {
                  if (pcFw2 != g_sFirmwareName2)
                  {
                     /* also store the new firmware binary name within global
                        configuration to be used for next download */
                     if (ldFw.data.pFirmware2 != DSL_NULL)
                     {
                        DSL_CPE_Free(g_sFirmwareName2);
                        g_sFirmwareName2 = DSL_CPE_Malloc (strlen (pcFw2) + 1);
                        if (g_sFirmwareName2)
                        {
                           strcpy (g_sFirmwareName2, pcFw2);
                        }
                     }
                  }
               }
            }
         }
#endif
        
         if ( ((ldFw.data.pFirmware  != DSL_NULL) && (ldFw.data.nFirmwareSize)) ||
              ((ldFw.data.pFirmware2 != DSL_NULL) && (ldFw.data.nFirmwareSize2)) )
         {
            ldFw.data.bLastChunk = DSL_TRUE;

            nRet = (DSL_Error_t) DSL_CPE_Ioctl(pContext->fd,
               DSL_FIO_AUTOBOOT_LOAD_FIRMWARE, (DSL_int_t) &ldFw);
         
            if (nRet < DSL_SUCCESS)
            {
               DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
                  "Autoboot Load Firmware failed!, nRet = %d!" DSL_CPE_CRLF,
                  ldFw.accessCtl.nReturn));
               break;
            }
         }
         else
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
               "Both FW binaries are not specified!" DSL_CPE_CRLF));
            nRet = DSL_ERROR;
            break;
         }
      }
   } while (0);

#ifdef DSL_CPE_SOAP_FW_UPDATE
   DSL_CPE_LockSet(&pContext->semFwUpdate);
#endif /* DSL_CPE_SOAP_FW_UPDATE */

   /* take care to clean up all memory that might be allocated during function
      processing */
   if (fd_image != DSL_NULL)
   {
      DSL_CPE_FClose(fd_image);
   }

   if (pChunkData != DSL_NULL)
   {
      DSL_CPE_Free(pChunkData);
      pChunkData = DSL_NULL;
      ldFw.data.pFirmware = DSL_NULL;
   }

   if (ldFw.data.pFirmware)
   {
      DSL_CPE_Free(ldFw.data.pFirmware);
   }

   if (ldFw.data.pFirmware2)
   {
      DSL_CPE_Free(ldFw.data.pFirmware2);
   }

   return nRet;
}

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
DSL_CPE_STATIC  DSL_char_t const * DSL_CPE_Event_Type2String(DSL_EventType_t *pEventType)
{
   DSL_int32_t i;
   DSL_CPE_EVT_CodeString_t const *pCS;

   pCS = eventString;
   for (i = 0; pCS[i].string; i++)
   {
      if (pCS[i].eventType == *pEventType)
      {
         return pCS[i].string;
      }
   }

   return "Unknown";
}
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
DSL_CPE_STATIC  DSL_boolean_t DSL_CPE_ScriptSectionTagFind(
   DSL_char_t *pLine,
   DSL_CPE_ScriptSection_t searchSection)
{
   DSL_boolean_t bTagOpenFound = DSL_FALSE, 
                 bTagCloseFound = DSL_FALSE,
                 bTagNameFound = DSL_FALSE,
                 bRet = DSL_FALSE;
   DSL_char_t *pTag = DSL_NULL;
   DSL_uint8_t nTagSize = 0;
   
   if (pLine == DSL_NULL)
   {
      return DSL_FALSE;
   }

   /* find tag open*/   
   while (*pLine != '\0')
   {
      if (*pLine == '[')
      {
         /* tag open found*/
         bTagOpenFound = DSL_TRUE;
         /* move to the next character*/
         pLine++;
         break;
      }
      /* move to the next character*/
      pLine++;
   }

   /* find tag name*/
   if (bTagOpenFound)
   {
      /* remove leading whitespaces*/
      while (isspace((DSL_int_t)*pLine) && *pLine != '\0')
      {
         pLine++;
      }
      
      if (searchSection == DSL_SCRIPT_SECTION_WAIT_FOR_CONFIFURATION)
      {
         pTag = DSL_CPE_SCRIPT_WAIT_FOR_CONFIGURATION_TAG;
         nTagSize = sizeof(DSL_CPE_SCRIPT_WAIT_FOR_CONFIGURATION_TAG)-1;
      }
      else if (searchSection == DSL_SCRIPT_SECTION_WAIT_FOR_LINK_ACTIVATE)
      {
         pTag = DSL_CPE_SCRIPT_WAIT_FOR_LINK_ACTIVATE_TAG;
         nTagSize = sizeof(DSL_CPE_SCRIPT_WAIT_FOR_LINK_ACTIVATE_TAG)-1;
      }
      else if (searchSection == DSL_SCRIPT_SECTION_COMMON)
      {
        pTag = DSL_CPE_SCRIPT_COMMON_TAG;
        nTagSize = sizeof(DSL_CPE_SCRIPT_COMMON_TAG)-1;
      }
      else if (searchSection == DSL_SCRIPT_SECTION_WAIT_BEFORE_RESTART)
      {
        pTag = DSL_CPE_SCRIPT_WAIT_BEFORE_RESTART_TAG;
        nTagSize = sizeof(DSL_CPE_SCRIPT_WAIT_BEFORE_RESTART_TAG)-1;
      }

      if (pTag)
      {
         /* compare case insensitive*/
         if (DSL_CPE_STRNCASECMP(pTag, pLine, nTagSize) == 0)
         {
            bTagNameFound = DSL_TRUE;
         }
      }
   }
   
   /* find tag close*/
   if (bTagNameFound)
   {
      /* remove ending whitespaces*/
      while (isspace((DSL_int_t)*pLine) && *pLine != '\0')
      {
         pLine++;
      }

      while (*pLine != '\0')
      {
         if (*pLine == ']')
         {
            /* tag open found*/
            bTagCloseFound = DSL_TRUE;
            break;
         }
         /* move to the next character*/
         pLine++;
      }
   }

   if (bTagOpenFound && bTagNameFound && bTagCloseFound && pTag)
   {
      bRet = DSL_TRUE;      
   }
   
   return bRet;
}

DSL_CPE_STATIC  DSL_boolean_t DSL_CPE_ScriptSectionOpenFind(
   DSL_char_t *pLine)
{
   DSL_boolean_t bRet = DSL_FALSE;
   
   if (pLine == DSL_NULL)
   {
      return DSL_FALSE;
   }

   /* find section open*/   
   while (*pLine != '\0')
   {
      if (*pLine == '{')
      {
         /* section open found*/
         bRet = DSL_TRUE;
         break;
      }
      /* move to the next character*/
      pLine++;
   }

   return bRet;
}

DSL_CPE_STATIC  DSL_boolean_t DSL_CPE_ScriptSectionCloseFind(
   DSL_char_t *pLine)
{
   DSL_boolean_t bRet = DSL_FALSE;
   
   if (pLine == DSL_NULL)
   {
      return DSL_FALSE;
   }

   /* find section close*/
   while (*pLine != '\0')
   {
      if (*pLine == '}')
      {
         /* section close found*/
         bRet = DSL_TRUE;
         break;
      }
      /* move to the next character*/
      pLine++;
   }

   return bRet;
}

/**
   Parses a batch file and performs necessary actions
*/
DSL_Error_t DSL_CPE_ScriptExecute (
   DSL_CPE_Control_Context_t *pContext,
   DSL_CPE_File_t *pFile,
   DSL_CPE_ScriptSection_t searchSection)
{
   DSL_char_t buf[128];
   DSL_char_t *pLine = DSL_NULL;
   DSL_char_t str_command[sizeof(buf)] = { 0 };
   DSL_boolean_t bTagFound = DSL_FALSE,
                 bSectionOpenFound = DSL_FALSE;
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
   DSL_char_t str_group[20] = { 0 };
   DSL_char_t str_value[128] = { 0 };
   DSL_char_t op1[40], op2[40];
   DSL_int_t address, index;
   DSL_int_t n = 0;
   DSL_uint16_t value = 0;
   DSL_uint16_t var16[8];
   DSL_uint32_t var32[8];
   DSL_uint16_t Message[DSL_MAX_CMV_MSG_LENGTH];
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_DeviceMessage_t msg;
#endif /* defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) ||
          defined (INCLUDE_DSL_CPE_API_AMAZON_S)*/

   if (!pFile)
   {
      return DSL_ERROR;
   }

   /* scan one line into buffer */
   while ((DSL_CPE_FGets (buf, sizeof(buf), pFile)) != DSL_NULL)
   {
      pLine = buf;
      /* if the line is empty or a special "#" symbol detected,
         then go on to the next */
      if ((sscanf (buf, "%s", &str_command) == 0) || (buf[0] == '#'))
      {
         continue;
      }

      /* remove leading whitespaces*/
      while((DSL_int_t)isspace(*pLine))
      {
         pLine++;
      }

      if (!bTagFound)
      {
         /* find search section tag*/
         if (!DSL_CPE_ScriptSectionTagFind(pLine, searchSection))
         {
            continue;            
         }
         else
         {
            /* search section tag found*/
            bTagFound = DSL_TRUE;
         }
      }

      /* find search section open*/
      if (bTagFound && !bSectionOpenFound)
      {
         if (DSL_CPE_ScriptSectionOpenFind(pLine))
         {
            /* search section open found*/
            bSectionOpenFound = DSL_TRUE;
         }
         continue; 
      }

      /* Check for search section close*/
      if (bSectionOpenFound)
      {
         if (DSL_CPE_ScriptSectionCloseFind(pLine))
         {
            /* Complete section proceeded, reset flags*/
            bTagFound = DSL_FALSE;
            bSectionOpenFound = DSL_FALSE;
            continue;
         }
      }

      if (strcmp (str_command, "echo") == 0)       /* echo command */
      {
         DSL_CPE_Echo (buf);
         memset (str_command, 0, sizeof(str_command));
         memset (buf, 0, sizeof(buf));
         continue;
      }

#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
      memset(&msg, 0, sizeof(DSL_DeviceMessage_t));

      /* if the line is empty, then go on to the next */
      if (strcmp (str_command, "cr") == 0)      /* cr command */
      {
         sscanf (buf, "%s %s %d %d %s", &str_command[0], &str_group[0], &address,
            &index, &str_value[0]);

         if (strncmp (str_value, "$", 1) != 0)
         {
            return DSL_ERR_INVALID_PARAMETR;
         }
         n = strtoul (str_value + 1, DSL_NULL, 0);
         nErrCode = DSL_CMV_Read (pContext, str_group, address, index, 1, &var16[n]);
         DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "read %s %d %d, value=%04x, "
            "retCode=%d\n", str_group, address, index, var16[n], nErrCode);
      }
      else if (strcmp (str_command, "cw") == 0) /* cw command */
      {
         sscanf (buf, "%s %s %d %d %s", &str_command[0], &str_group[0], &address,
            &index, &str_value[0]);

         n = strtoul (str_value + 1, DSL_NULL, 0);
         if (strncmp (str_value, "$", 1) == 0)
         {
            value = var16[n];
         }
         else if (strncmp (str_value, "#", 1) == 0)
         {
            value = strtoul (str_value + 1, DSL_NULL, 0);
         }
         else
         {
            value = strtoul (str_value, DSL_NULL, 0);
         }
         if (DSL_CMV_Write (pContext, str_group, address, index, 1, &value) != DSL_SUCCESS)
         {
            return DSL_ERR_MSG_EXCHANGE;
         }
      }
      else if (strcmp (str_command, "mw") == 0) /* debug memory write command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &str_value[0]);

         n = strtoul (str_value + 1, DSL_NULL, 0);
         address = strtoul (op1, DSL_NULL, 0);
         if (strncmp (str_value, "$", 1) == 0)
         {
            value = var16[n];
         }
         else if (strncmp (str_value, "#", 1) == 0)
         {
            value = strtoul (str_value + 1, DSL_NULL, 0);
         }
         else
         {
            value = strtoul (str_value, DSL_NULL, 0);
         }

         DSL_CMV_Prepare (DSL_CMV_OPCODE_H2D_DEBUG_WRITE_DM, 0x0, address >> 16,
            (address) & (0xffff), 1, &value, Message);

         msg.data.pMsg = (DSL_uint8_t *)Message;
         msg.data.nSizeTx = 8;
         if (DSL_CPE_Ioctl (pContext->fd, DSL_FIO_DBG_DEVICE_MESSAGE_SEND, (int)&Message) < 0)
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "mw %08x fail", address);
            return DSL_ERR_MSG_EXCHANGE;
         }
      }
      else if (strcmp (str_command, "mr") == 0) /* debug memory read command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &str_value[0]);
         address = strtoul (op1, DSL_NULL, 0);
         n = strtoul (str_value + 1, DSL_NULL, 0);
         if (strncmp (str_value, "$", 1) != 0)
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "error:mr parameter type "
               "mismatch!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
         DSL_CMV_Prepare (DSL_CMV_OPCODE_H2D_DEBUG_READ_DM, 0x0, address >> 16,
            (address) & (0xffff), 1, DSL_NULL, Message);

         msg.data.pMsg = (DSL_uint8_t *)Message;
         msg.data.nSizeTx = 8;
         msg.data.nSizeRx = 10;
         if (DSL_CPE_Ioctl (pContext->fd, DSL_FIO_DBG_DEVICE_MESSAGE_SEND, (int)&Message) < 0)
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "mw %08x fail", address);
            return DSL_ERR_MSG_EXCHANGE;
         }
         value = Message[4];
         var16[n] = value;
      }


      /* meiw command */
/*$$
      else if (strcmp (str_command, "meiw") == 0)
      {
         meireg regrdwr;

         sscanf (buf, "%s %s %s", &str_command, &op1, &op2);
         regrdwr.iAddress = strtoul (op1, DSL_NULL, 0) + MEI_SPACE_ACCESS;


         if (strncmp (op2, "#", 1) == 0)
            regrdwr.iData = strtoul (op1 + 1, DSL_NULL, 0);
         else if (strncmp (op2, "$", 1) == 0)
            regrdwr.iData = var16[strtoul (op2 + 1, DSL_NULL, 0)];
         else if (strncmp (op2, "@", 1) == 0)
            regrdwr.iData = var32[strtoul (op2 + 1, DSL_NULL, 0)];
         else
            regrdwr.iData = strtoul (op1, DSL_NULL, 0);
         // printf("address=%08x\n", regrdwr.iAddress);

         msg.data.pMsg = (DSL_uint8_t *)Message;
         msg.data.nSizeTx = 8;
         msg.data.nSizeRx = 10;

         if (ioctl (pContext->fd, DANUBE_MEI_CMV_WRITE, &regrdwr) < 0)
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "meiw fail\n");
            break;;
         }

      }
*/

      /* meir command */
/*$$
      else if (strcmp (str_command, "meir") == 0)
      {
         meireg regrdwr;

         sscanf (buf, "%s %s %s", &str_command, &op1, &op2);
         regrdwr.iAddress = strtoul (op1, DSL_NULL, 0) + MEI_SPACE_ACCESS;
         if (ioctl (pContext->fd, DANUBE_MEI_CMV_READ, &regrdwr) < 0)
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "meir fail\n");
            break;
         }
         if (strncmp (op2, "$", 1) == 0)
            var16[strtoul (op2 + 1, DSL_NULL, 0)] = regrdwr.iData;
         else if (strncmp (op2, "@", 1) == 0)
            var32[strtoul (op2 + 1, DSL_NULL, 0)] = regrdwr.iData;
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "meir grammar error!" DSL_CPE_CRLF);
            break;
         }

      }
*/

      else if (strcmp (str_command, "lst") == 0)        /* list command */
      {
         sscanf (buf, "%s %s", &str_command[0], &str_value[0]);

         if (strncmp (str_value, "$", 1) == 0)
         {
            n = strtoul (str_value + 1, DSL_NULL, 0);
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "$%d=0x%04x\n", n, var16[n]);
         }
         else if (strncmp (str_value, "@", 1) == 0)
         {
            n = strtoul (str_value + 1, DSL_NULL, 0);
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "$%d=0x%08x\n", n, var32[n]);
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "lst grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else if (strcmp (str_command, "mov") == 0)        /* move command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] =
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] =
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] = strtoul (op2 + 1, DSL_NULL, 0);
            }
            else
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] = strtoul (op2, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] =
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] =
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] = strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else if (strcmp (str_command, "or") == 0) /* or command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] |=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] |=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] |= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] |=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] |=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] |= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }

      else if (strcmp (str_command, "and") == 0)        /* and command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] &=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] &=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] &= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] &=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] &=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] &= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else if (strcmp (str_command, "not") == 0)        /* not command */
      {
         sscanf (buf, "%s %s", &str_command[0], &op1[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            var16[strtoul (op1 + 1, DSL_NULL, 0)] =
               ~var16[strtoul (op1 + 1, DSL_NULL, 0)];
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            var32[strtoul (op1 + 1, DSL_NULL, 0)] =
               ~var32[strtoul (op1 + 1, DSL_NULL, 0)];
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else if (strcmp (str_command, "shl") == 0)        /* shift left command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] <<= var16[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] <<= var32[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] <<= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] <<= var16[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] <<= var32[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] <<= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }

      else if (strcmp (str_command, "shr") == 0)        /* shift right command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] >>= var16[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] >>= var32[strtoul (op2 + 1,
                     DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] >>= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] >>= var16[strtoul (op2 + 1,
                     DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] >>= var32[strtoul (op2 + 1,
                     DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] >>= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }

      else if (strcmp (str_command, "add") == 0)        /* addition command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] +=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] +=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] += strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] +=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] +=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] += strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else if (strcmp (str_command, "sub") == 0)        /* subtraction command */
      {
         sscanf (buf, "%s %s %s", &str_command[0], &op1[0], &op2[0]);
         if (strncmp (op1, "$", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] -=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] -=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var16[strtoul (op1 + 1, DSL_NULL, 0)] -= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else if (strncmp (op1, "@", 1) == 0)
         {
            if (strncmp (op2, "$", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] -=
                  var16[strtoul (op2 + 1, DSL_NULL, 0)];
            }
            else if (strncmp (op2, "@", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] -=
                  var32[strtoul (op2 + 1, DSL_NULL, 0)];

            }
            else if (strncmp (op2, "#", 1) == 0)
            {
               var32[strtoul (op1 + 1, DSL_NULL, 0)] -= strtoul (op2 + 1, DSL_NULL, 0);
            }
         }
         else
         {
            DSL_CPE_FPrintf (DSL_CPE_STDOUT, DSL_CPE_PREFIX "grammar error!" DSL_CPE_CRLF);
            return DSL_ERR_INVALID_PARAMETR;
         }
      }
      else
#endif /* defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || 
          defined (INCLUDE_DSL_CPE_API_AMAZON_S)*/
      {
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
         /* Check if a CLI command is given */
         DSL_CPE_CLI_CommandExecute (pContext->fd, str_command,
            buf + strlen(str_command) + 1, DSL_CPE_STDOUT);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT*/
      }

      memset (str_command, 0, sizeof(str_command));
      memset (buf, 0, sizeof(buf));
   }
   return DSL_SUCCESS;
}
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */

#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
/**
   Performs an execution of debug batch file
*/
DSL_Error_t DSL_CPE_ScriptFileParse(
   DSL_CPE_Control_Context_t * pContext,
   DSL_char_t *sFileName,
   DSL_CPE_ScriptSection_t searchSection)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;
   DSL_CPE_File_t *pFile;

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX "Parsing script file (%s)..."
      DSL_CPE_CRLF, sFileName));

   do
   {
      if (sFileName == DSL_NULL)
      {
         nErrCode = DSL_ERR_POINTER;
         break;
      }

      pFile = DSL_CPE_FOpen(sFileName, "r");
      if (pFile == DSL_NULL)
      {
         nErrCode = DSL_ERROR;
         break;
      }

      nErrCode = DSL_CPE_ScriptExecute(pContext, pFile, searchSection);

      DSL_CPE_FClose(pFile);

   } while (0);


   if (nErrCode != DSL_SUCCESS)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "Script file execution failed (nErrCode=%d)!" DSL_CPE_CRLF, nErrCode));
   }
   else
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX
         "Script file execution performed successfully." DSL_CPE_CRLF));
   }

   return nErrCode;
}
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */

#ifdef INCLUDE_SCRIPT_NOTIFICATION
DSL_CPE_STATIC  DSL_void_t DSL_CPE_ScriptRun()
{
   if (g_sRcScript != DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX "Run script file (%s)"
         DSL_CPE_CRLF, g_sRcScript));

      if (DSL_CPE_System(g_sRcScript) != DSL_SUCCESS
         && bScriptWarn == DSL_FALSE)
      {
         bScriptWarn = DSL_TRUE;
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "notification script file (%s) execution failed!" DSL_CPE_CRLF,
               g_sRcScript));
      }
   }

   return;
}
#endif /* #ifdef INCLUDE_SCRIPT_NOTIFICATION */

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_InitReadyHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s " DSL_CPE_CRLF ,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)));
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC DSL_int_t DSL_CPE_Event_S_AutobootStatusHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventStatus_t *pEvent)
{
#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   DSL_CPE_ScriptSection_t searchSection;
   DSL_AutobootStatus_t status;
   DSL_AutobootStatusData_t StatusData;
   DSL_AutobootControl_t control;
   DSL_int_t ret = 0;
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */

   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nStatus=%d nFirmwareRequestType=%d" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.pData->autobootStatus.nStatus, pEvent->data.pData->autobootStatus.nFirmwareRequestType);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   memcpy(&StatusData, pEvent->data.pData, sizeof(DSL_AutobootStatusData_t));
   
   switch (StatusData.nStatus)
   {
   case DSL_AUTOBOOT_STATUS_CONFIG_WRITE_WAIT:
      searchSection = DSL_SCRIPT_SECTION_WAIT_FOR_CONFIFURATION;
      break;
   case DSL_AUTOBOOT_STATUS_LINK_ACTIVATE_WAIT:
      searchSection = DSL_SCRIPT_SECTION_WAIT_FOR_LINK_ACTIVATE;
      break;
   case DSL_AUTOBOOT_STATUS_RESTART_WAIT:
      searchSection = DSL_SCRIPT_SECTION_WAIT_BEFORE_RESTART;
      break;
   default:
      return 0;
   }
            
   if (pContext->nFwReqType == DSL_FW_REQUEST_ADSL)
   {
      if (g_sAdslScript != DSL_NULL)
      {
         DSL_CPE_ScriptFileParse(pContext, g_sAdslScript, searchSection);
      }
      else
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX
            "ADSL script not specified" DSL_CPE_CRLF));
      }
   }

#ifdef INCLUDE_DSL_CPE_API_VINAX
   if (pContext->nFwReqType == DSL_FW_REQUEST_VDSL)
   {
      if (g_sVdslScript != DSL_NULL)
      {
         DSL_CPE_ScriptFileParse(pContext, g_sVdslScript, searchSection);
      }
      else
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX
            "VDSL script not specified" DSL_CPE_CRLF));
      }
   }
#endif /* INCLUDE_DSL_CPE_API_VINAX */
      
   /* Get autoboot status directly after script execution*/
   memset (&status, 0, sizeof(DSL_AutobootStatus_t));

   ret = DSL_CPE_Ioctl(pContext->fd, DSL_FIO_AUTOBOOT_STATUS_GET,(DSL_int_t) &status);

   if (ret >= 0)
   {
      if (((status.data.nStatus == DSL_AUTOBOOT_STATUS_CONFIG_WRITE_WAIT)  && g_bAutoContinueWaitBeforeConfigWrite) ||
          ((status.data.nStatus == DSL_AUTOBOOT_STATUS_LINK_ACTIVATE_WAIT) && g_bAutoContinueWaitBeforeLinkActivation) ||
          ((status.data.nStatus == DSL_AUTOBOOT_STATUS_RESTART_WAIT) && g_bAutoContinueWaitBeforeRestart))
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_WRN,
            (DSL_CPE_PREFIX "Continue autoboot handling..." DSL_CPE_CRLF));

         memset (&control, 0, sizeof(DSL_AutobootControl_t));

         control.data.nCommand = DSL_AUTOBOOT_CTRL_CONTINUE;

         ret = DSL_CPE_Ioctl (pContext->fd, DSL_FIO_AUTOBOOT_CONTROL_SET,(DSL_int_t) &control);
      }
   }
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */

   return 0;
}

#ifdef INCLUDE_DSL_PM
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
DSL_CPE_STATIC DSL_int_t DSL_CPE_Event_S_PmSyncHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s b15MinElapsed=%u b1DayElapsed=%u" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.pData->sync.b15MinElapsed,
      pEvent->data.pData->sync.b1DayElapsed);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}
#endif /*INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /*INCLUDE_DSL_PM*/

DSL_CPE_STATIC DSL_int_t DSL_CPE_Event_S_LineStateHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventStatus_t *pEvent)
{
   DSL_LineStateValue_t nLineState = DSL_LINESTATE_NOT_INITIALIZED;
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   DSL_boolean_t bExec = DSL_FALSE;
#endif

   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

   nLineState = pEvent->data.pData->lineStateData.nLineState;

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s " "nLineState=%08X" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.pData->lineStateData.nLineState);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

#ifdef INCLUDE_SCRIPT_NOTIFICATION
   if (g_sRcScript != DSL_NULL)
   {
      if ( (nLineState == DSL_LINESTATE_SHOWTIME_TC_SYNC) &&
         (g_nPrevLineState != DSL_LINESTATE_SHOWTIME_TC_SYNC) )
      {
         if (DSL_CPE_SetEnv("DSL_INTERFACE_STATUS", "UP") == DSL_SUCCESS)
         {
            bExec = DSL_TRUE;
         }
      }
      else if ( (nLineState != DSL_LINESTATE_SHOWTIME_TC_SYNC) &&
         (g_nPrevLineState == DSL_LINESTATE_SHOWTIME_TC_SYNC) )
      {
         if (DSL_CPE_SetEnv("DSL_INTERFACE_STATUS", "DOWN") == DSL_SUCCESS)
         {
            bExec = DSL_TRUE;
         }
      }

      if (bExec != DSL_FALSE)
      {
         if (DSL_CPE_SetEnv("DSL_NOTIFICATION_TYPE", "DSL_INTERFACE_STATUS") == DSL_SUCCESS)
         {
            DSL_CPE_ScriptRun();
         }
      }
   }
#endif

   g_nPrevLineState = nLineState;

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_SystemInterfaceStatusHandle(
   DSL_EventStatus_t *pEvent)
{
   DSL_Error_t nErrCode = DSL_SUCCESS;

   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_SCRIPT_NOTIFICATION
   if (g_sRcScript != DSL_NULL)
   {
      if (DSL_CPE_SetEnv("DSL_NOTIFICATION_TYPE", "DSL_STATUS") == DSL_SUCCESS)
      {
         nErrCode = DSL_CPE_SetEnv("DSL_XTU_STATUS", "ADSL");
         if (nErrCode == DSL_SUCCESS)
         {
            switch (pEvent->data.pData->systemInterfaceStatusData.nTcLayer)
            {
               case DSL_TC_ATM:
                  nErrCode = DSL_CPE_SetEnv("DSL_TC_LAYER_STATUS", "ATM");
                  break;
               case DSL_TC_EFM:
               case DSL_TC_EFM_FORCED:
                  nErrCode = DSL_CPE_SetEnv("DSL_TC_LAYER_STATUS", "EFM");
                  break;
               default:
                  nErrCode = DSL_CPE_SetEnv("DSL_TC_LAYER_STATUS", "UNKNOWN");
                  break;
            }
         }

         if (nErrCode == DSL_SUCCESS)
         {
            switch (pEvent->data.pData->systemInterfaceStatusData.nEfmTcConfigUs)
            {
               case DSL_EMF_TC_NORMAL:
                  nErrCode = DSL_CPE_SetEnv("DSL_EFM_TC_CONFIG_US", "NORMAL");
                  break;
               case DSL_EMF_TC_PRE_EMPTION:
                  nErrCode = DSL_CPE_SetEnv("DSL_EFM_TC_CONFIG_US", "PRE_EMPTION");
                  break;
               default:
                  nErrCode = DSL_CPE_SetEnv("DSL_EFM_TC_CONFIG_US", "UNKNOWN");
                  break;
            }
         }

         if (nErrCode == DSL_SUCCESS)
         {
            switch (pEvent->data.pData->systemInterfaceStatusData.nEfmTcConfigDs)
            {
               case DSL_EMF_TC_NORMAL:
                  nErrCode = DSL_CPE_SetEnv("DSL_EFM_TC_CONFIG_DS", "NORMAL");
                  break;
               default:
                  nErrCode = DSL_CPE_SetEnv("DSL_EFM_TC_CONFIG_US", "UNKNOWN");
                  break;
            }
         }

         if (nErrCode == DSL_SUCCESS)
         {
            DSL_CPE_ScriptRun();
         }
      }
   }
#endif

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nTcLayer=%d nEfmTcConfigUs=0x%08X nEfmTcConfigDs=0x%08X "
      "nSystemIf=%d" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.pData->systemInterfaceStatusData.nTcLayer,
      pEvent->data.pData->systemInterfaceStatusData.nEfmTcConfigUs,
      pEvent->data.pData->systemInterfaceStatusData.nEfmTcConfigDs,
      pEvent->data.pData->systemInterfaceStatusData.nSystemIf);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return nErrCode;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_I_DataPathFailuresHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nChannel=%d nAccessDir=%s nDataPathFailures=%08X" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nAccessDir == DSL_DOWNSTREAM ? "DSL_DOWNSTREAM" : "DSL_UPSTREAM",
      pEvent->data.pData->dataPathFailuresData.nDataPathFailures);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_I_LineFailuresHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nChannel=%d nAccessDir=%s nLineFailures=%08X" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nAccessDir == DSL_DOWNSTREAM ? "DSL_DOWNSTREAM" : "DSL_UPSTREAM",
      pEvent->data.pData->lineFailuresData.nLineFailures);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_ChannelDataRateHandle(
   DSL_EventStatus_t *pEvent)
{
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   DSL_char_t sVarName[20];
   DSL_char_t sVarVal[11];
#endif
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL)
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_SCRIPT_NOTIFICATION
   if (g_sRcScript != DSL_NULL)
   {
      if (DSL_CPE_SetEnv("DSL_NOTIFICATION_TYPE", "DSL_DATARATE_STATUS") == DSL_SUCCESS)
      {
         sprintf(sVarName, "DSL_DATARATE_%s_BC%d",
            pEvent->data.nAccessDir == DSL_DOWNSTREAM ? "DS" : "US",
            pEvent->data.nChannel == 0 ? 0 : pEvent->data.nChannel == 1 ? 1 : 0);
         sprintf(sVarVal, "%lu", pEvent->data.pData->channelStatusData.ActualDataRate);

         if (DSL_CPE_SetEnv(sVarName, sVarVal) == DSL_SUCCESS)
         {
            DSL_CPE_ScriptRun();
         }
      }
   }
#endif

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nChannel=%d nAccessDir=%s ActualDataRate=%d PreviousDataRate=%d "
      "ActualInterleaveDelay=%d ActualImpulseNoiseProtection=%d" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nAccessDir == DSL_DOWNSTREAM ? "DSL_DOWNSTREAM" : "DSL_UPSTREAM",
   pEvent->data.pData->channelStatusData.ActualDataRate,
   pEvent->data.pData->channelStatusData.PreviousDataRate,
   pEvent->data.pData->channelStatusData.ActualInterleaveDelay,
   pEvent->data.pData->channelStatusData.ActualImpulseNoiseProtection);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_ChannelDataRateShiftThresholdCrossingHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nChannel=%d nAccessDir=%s"
      " nDataRateThresholdType=DSL_G997_DATARATE_THRESHOLD_%sSHIFT" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nAccessDir == DSL_DOWNSTREAM ? "DSL_DOWNSTREAM" : "DSL_UPSTREAM",
      pEvent->data.pData->dataRateThresholdCrossing.nDataRateThresholdType ==
         DSL_G997_DATARATE_THRESHOLD_DOWNSHIFT ? "DOWN": "UP");
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_LinitFailureHandle(
   DSL_EventStatus_t *pEvent)
{
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_char_t *pStatus = DSL_NULL;
#endif

   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   switch(pEvent->data.pData->lineInitStatusData)
   {
   case LINIT_SUCCESSFUL:
      pStatus = "LINIT_SUCCESSFUL";
      break;
   case LINIT_CONFIG_ERROR:
      pStatus = "LINIT_CONFIG_ERROR";
      break;
   case LINIT_CONFIG_NOT_FEASIBLE:
      pStatus = "LINIT_CONFIG_NOT_FEASIBLE";
      break;
   case LINIT_COMMUNICATION_PROBLEM:
      pStatus = "LINIT_COMMUNICATION_PROBLEM";
      break;
   case LINIT_NO_PEER_XTU:
      pStatus = "LINIT_NO_PEER_XTU";
      break;
   case LINIT_UNKNOWN:
      pStatus = "LINIT_UNKNOWN";
      break;

   default:
      pStatus = "LINIT_UNKNOWN";
      break;
   }

   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s lineInitStatusData=%s" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)), pStatus);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_FirmwareRequestHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventType_t nEvent,
   DSL_FirmwareRequestType_t nFwReqType)
{
   DSL_Error_t nRet = 0;

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_char_t *pFwReqType = "DSL_FW_REQUEST_NA";
   
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "sEventType=%s ",
      DSL_CPE_Event_Type2String(&nEvent));

   switch (nFwReqType)
   {
      case DSL_FW_REQUEST_ADSL:
         pFwReqType = "DSL_FW_REQUEST_ADSL";
         break;
      case DSL_FW_REQUEST_VDSL:
         pFwReqType = "DSL_FW_REQUEST_VDSL";
         break;
      case DSL_FW_REQUEST_NA:
      default:
         pFwReqType = "DSL_FW_REQUEST_NA";
         break;
   }

   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "sFirmwareRequestType=%s"DSL_CPE_CRLF,
      pFwReqType);

#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   nRet = DSL_CPE_DownloadFirmware(nFwReqType, DSL_NULL, DSL_NULL);

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_CRLF));

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_FirmwareDownloadStatusHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventType_t nEvent,
   DSL_FwDownloadStatusData_t nFwDwnlStatus)
{
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "sEventType=%s nError=%d nFwType=%d",
      DSL_CPE_Event_Type2String(&nEvent),
      nFwDwnlStatus.nError,
      nFwDwnlStatus.nFwType);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   if (nFwDwnlStatus.nError == DSL_FW_LOAD_SUCCESS)
   {
      pContext->nFwReqType = nFwDwnlStatus.nFwType;
   }
   else
   {
      pContext->nFwReqType = DSL_FW_REQUEST_NA;
   }
#endif
#endif

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_CRLF));

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_FeInventoryAvailableHandle(
   DSL_CPE_Control_Context_t *pContext,
   DSL_EventStatus_t *pEvent)
{
   DSL_G997_LineInventory_t lineInventory;
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_char_t buf[256];
#endif
   DSL_int_t ret = 0;

   if( pEvent == DSL_NULL )
   {
      return DSL_ERROR;
   }

   memset(&lineInventory, 0x00, sizeof(DSL_G997_LineInventory_t));
   /** Specify direction*/
   lineInventory.nDirection = DSL_FAR_END;
   ret = DSL_CPE_Ioctl(pContext->fd, DSL_FIO_G997_LINE_INVENTORY_GET,
      (DSL_int_t) &lineInventory);

   if( ret < 0 )
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "FE line inventory get failed!, ret = %d!"DSL_CPE_CRLF,
         lineInventory.accessCtl.nReturn));
      return -1;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "sEventType=%s ",
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)));

   DSL_CPE_ArraySPrintF(buf, lineInventory.data.G994VendorID,
      sizeof(lineInventory.data.G994VendorID),
      sizeof(lineInventory.data.G994VendorID[0]),
      DSL_ARRAY_FORMAT_HEX);
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sG994VendorID=%s ",
      CLI_EventText, buf);

   DSL_CPE_ArraySPrintF(buf, lineInventory.data.SystemVendorID,
      sizeof(lineInventory.data.SystemVendorID),
      sizeof(lineInventory.data.SystemVendorID[0]),
      DSL_ARRAY_FORMAT_HEX);
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sSystemVendorID=%s ",
      CLI_EventText, buf);

   DSL_CPE_ArraySPrintF(buf, lineInventory.data.VersionNumber,
      sizeof(lineInventory.data.VersionNumber),
      sizeof(lineInventory.data.VersionNumber[0]),
      DSL_ARRAY_FORMAT_HEX);
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sVersionNumber=%s ",
      CLI_EventText, buf);

   DSL_CPE_ArraySPrintF(buf, lineInventory.data.SerialNumber,
      sizeof(lineInventory.data.SerialNumber),
      sizeof(lineInventory.data.SerialNumber[0]),
      DSL_ARRAY_FORMAT_HEX);
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sSerialNumber=%s ",
      CLI_EventText, buf);

   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sSelfTestResult=%u ",
      CLI_EventText, lineInventory.data.SelfTestResult);

   DSL_CPE_ArraySPrintF(buf, lineInventory.data.XTSECapabilities,
      sizeof(lineInventory.data.XTSECapabilities),
      sizeof(lineInventory.data.XTSECapabilities[0]),
      DSL_ARRAY_FORMAT_HEX);
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%sXTSECapabilities=%s ",
      CLI_EventText, buf);

   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%s" DSL_CPE_CRLF, CLI_EventText);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_LinePowerMgmtStateHandle(
   DSL_EventStatus_t *pEvent)
{
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_char_t *pStatus = DSL_NULL;
#endif

   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   switch(pEvent->data.pData->powerMgmtStatusData.nPowerManagementStatus)
   {
   case DSL_G997_PMS_L0:
      pStatus = "DSL_G997_PMS_L0";
      break;
   case DSL_G997_PMS_L1:
      pStatus = "DSL_G997_PMS_L1";
      break;
   case DSL_G997_PMS_L2:
      pStatus = "DSL_G997_PMS_L2";
      break;
   case DSL_G997_PMS_L3:
      pStatus = "DSL_G997_PMS_L3";
      break;

   default:
      pStatus = "DSL_G997_PMS_NA";
      break;
   }

   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s nPowerManagementStatus=%s" DSL_CPE_CRLF,
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)), pStatus);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

#if defined(INCLUDE_DSL_CEOC)
DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_S_SnmpMessageAvailableHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s",
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)));
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}
#endif

#if defined(INCLUDE_DSL_PM)
#if defined(INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS)
DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_I_ChannelThresholdCrossingHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s "
      "nChannel=%d "
      "nATUDir=%s "
      "nCurr15MinTime=%d "
      "nCurr1DayTime=%d "
      "n15Min=0x%08X "
      "n1day=0x%08X ",
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nXtuDir == DSL_NEAR_END ? "DSL_NEAR_END" : "DSL_FAR_END",
      pEvent->data.pData->channelThresholdCrossing.nCurr15MinTime,
      pEvent->data.pData->channelThresholdCrossing.nCurr1DayTime,
      pEvent->data.pData->channelThresholdCrossing.n15Min,
      pEvent->data.pData->channelThresholdCrossing.n1Day);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}
#endif /** #if defined(INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS)*/

DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_I_LineThresholdCrossingHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s "
      "nChannel=%d "
      "nATUDir=%s "
      "nCurr15MinTime=%d "
      "nCurr1DayTime=%d "
      "n15Min=0x%08X "
      "n1day=0x%08X ",
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nXtuDir == DSL_NEAR_END ? "DSL_NEAR_END" : "DSL_FAR_END",
      pEvent->data.pData->lineThresholdCrossing.nCurr15MinTime,
      pEvent->data.pData->lineThresholdCrossing.nCurr1DayTime,
      pEvent->data.pData->lineThresholdCrossing.n15Min,
      pEvent->data.pData->lineThresholdCrossing.n1Day);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}

#if defined(INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS)
DSL_CPE_STATIC  DSL_int_t DSL_CPE_Event_I_DataPathThresholdCrossingHandle(
   DSL_EventStatus_t *pEvent)
{
   if( pEvent == DSL_NULL || pEvent->data.pData == DSL_NULL )
   {
      return DSL_ERROR;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
      "sEventType=%s "
      "nChannel=%d "
      "nATUDir=%s "
      "nCurr15MinTime=%d "
      "nCurr1DayTime=%d "
      "n15Min=0x%08X "
      "n1day=0x%08X ",
      DSL_CPE_Event_Type2String(&(pEvent->data.nEventType)),
      pEvent->data.nChannel,
      pEvent->data.nXtuDir == DSL_NEAR_END ? "DSL_NEAR_END" : "DSL_FAR_END",
      pEvent->data.pData->dataPathThresholdCrossing.nCurr15MinTime,
      pEvent->data.pData->dataPathThresholdCrossing.nCurr1DayTime,
      pEvent->data.pData->dataPathThresholdCrossing.n15Min,
      pEvent->data.pData->dataPathThresholdCrossing.n1Day);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   return 0;
}
#endif /** #if defined(INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS)*/
#endif /** #if defined(INCLUDE_DSL_PM)*/

DSL_int_t DSL_CPE_EventHandler (DSL_CPE_Thread_Params_t *param)
{
   DSL_CPE_Control_Context_t *pContext;
#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
   DSL_CPE_File_t *file = DSL_NULL;
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */
   DSL_int_t num_open_fds = 0,
      sretval;
   DSL_CPE_fd_set_t rd_fds;               /* File descr. sets for use with select() */
   DSL_int_t ret;
   DSL_EventStatus_t event;
   DSL_ShowtimeLogging_t showtimeLogging;
   DSL_uint16_t showtimeData[120];
#if defined (INCLUDE_DSL_CPE_CLI_SUPPORT) || defined (INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT)
   DSL_int_t i;
#endif

/*   DSL_EventData_Union_t eventData;*/

   pContext = (DSL_CPE_Control_Context_t *) param->nArg1;
   if (pContext == DSL_NULL)
   {
      return DSL_ERROR;
   }

   num_open_fds = pContext->fd + 1;

   /*
    * The main fd read cycle starts here.
    */
   pContext->bEvtRun = DSL_TRUE;
   for (;;)
   {
      if (pContext->bRun == DSL_FALSE || pContext->bEvtRun == DSL_FALSE)
      {
         break;
      }

      FD_ZERO (&rd_fds);
      DSL_CPE_FD_SET (pContext->fd, &rd_fds);

      sretval = DSL_CPE_Select (num_open_fds, &rd_fds, NULL, DSL_DEV_TIMEOUT_SELECT);

      /* error or timeout on select */
      if (sretval <= 0)
      {
         continue;
      }

      /* handles a normal wakeup of the select because data is avialble from
         within the device driver */
      if (DSL_CPE_FD_ISSET (pContext->fd, &rd_fds) == 0)
      {
         continue;
      }

      DSL_CCA_DEBUG(DSL_CCA_DBG_MSG, (DSL_CPE_PREFIX "EVENT..." DSL_CPE_CRLF));

      /* get an event status */
      memset (&event, 0, sizeof (DSL_EventStatus_t));
      event.data.pData = (DSL_EventData_Union_t*)malloc(sizeof(DSL_EventData_Union_t));

      if( event.data.pData == DSL_NULL )
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX DSL_CPE_CRLF
            "Can't allocate memory for event data!!!"DSL_CPE_CRLF));
         continue;
      }

      ret =
         DSL_CPE_Ioctl (pContext->fd, DSL_FIO_EVENT_STATUS_GET,
         (DSL_int_t) & event);

      if (ret < 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
            "Could not get device driver event status, ret = %d!" DSL_CPE_CRLF, ret));
         if (event.data.pData != DSL_NULL)
         {
            free(event.data.pData);
         }
         continue;
      }

      if (event.accessCtl.nReturn != DSL_SUCCESS)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
            "Error code was returned by driver IOCTL errCode= %d!" DSL_CPE_CRLF,
            event.accessCtl.nReturn));
         continue;
      }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
      CLI_EventText[0] = '\0';
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

      switch (event.data.nEventType)
      {
      case DSL_EVENT_S_INIT_READY:
         DSL_CPE_Event_S_InitReadyHandle(&event);
         break;
      case DSL_EVENT_S_SHOWTIME_LOGGING:
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
         DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText),
            "sEventType=%s ",
            DSL_CPE_Event_Type2String(&(event.data.nEventType)));
#endif
         memset (&showtimeLogging, 0, sizeof (DSL_ShowtimeLogging_t));
         memset (&showtimeData, 0, sizeof (showtimeData));
         showtimeLogging.data.nDataSize = sizeof (showtimeData);
         showtimeLogging.data.pData = (DSL_uint8_t *) showtimeData;

         ret =
            DSL_CPE_Ioctl (pContext->fd, DSL_FIO_SHOWTIME_LOGGING_DATA_GET,
            (DSL_int_t) & showtimeLogging);
         if (ret < 0)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
               "Could not get showtime info, ret = %d!" DSL_CPE_CRLF, ret));
            continue;
         }

         if (showtimeLogging.accessCtl.nReturn != DSL_SUCCESS)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
               "Error code was returned by driver IOCTL while trying to get "
               "showtimeLogging data, errCode = %d!" DSL_CPE_CRLF,
               showtimeLogging.accessCtl.nReturn));
            continue;
         }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
         DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%s nData=\"",
            CLI_EventText);

         for (i = 0; i < (DSL_int_t)(sizeof (showtimeData) / 2); i++)
         {
            if (i % 10 == 0)
            {
               DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%s" DSL_CPE_CRLF,
                  CLI_EventText);
            }
            DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%s0x%04X ",
               CLI_EventText, showtimeData[i]);
         }
         DSL_CPE_snprintf(CLI_EventText, sizeof(CLI_EventText), "%s" DSL_CPE_CRLF "\"",
            CLI_EventText);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
         file = DSL_CPE_FOpen(DSL_CPE_SHOWTIME_EVENT_LOGGING_FILENAME,"w");
         if (file == DSL_NULL)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX DSL_CPE_CRLF
               "Could not open file (%s) for showtime event logging data!!!"
               DSL_CPE_CRLF, DSL_CPE_SHOWTIME_EVENT_LOGGING_FILENAME));
         }
         else
         {
            DSL_CPE_FPrintf(file, "showtime event logging data:");
            for (i = 0; i < (DSL_int_t)(sizeof (showtimeData) / 2); i++)
            {
               if (i % 10 == 0) DSL_CPE_FPrintf (file, DSL_CPE_CRLF);
               DSL_CPE_FPrintf (file, "0x%04X ", showtimeData[i]);
            }
            DSL_CPE_FPrintf (file, DSL_CPE_CRLF DSL_CPE_CRLF);

            DSL_CPE_FClose(file);
         }
#endif
         break;
      case DSL_EVENT_S_FIRMWARE_REQUEST:
         DSL_CPE_Event_S_FirmwareRequestHandle(pContext,
            event.data.nEventType,
            event.data.pData->fwRequestData.nFirmwareRequestType);
         break;
      case DSL_EVENT_S_FIRMWARE_DOWNLOAD_STATUS:
         DSL_CPE_Event_S_FirmwareDownloadStatusHandle(
            pContext,
            event.data.nEventType,
            event.data.pData->fwDownloadStatus);
         break;
      case DSL_EVENT_I_DATA_PATH_FAILURES:
         DSL_CPE_Event_I_DataPathFailuresHandle(&event);
         break;
      case DSL_EVENT_I_LINE_FAILURES:
         DSL_CPE_Event_I_LineFailuresHandle(&event);
         break;
      case DSL_EVENT_S_LINE_STATE:
         DSL_CPE_Event_S_LineStateHandle(pContext, &event);
         break;
#ifdef INCLUDE_DSL_PM
#ifdef INCLUDE_DSL_CPE_PM_HISTORY
      case DSL_EVENT_S_PM_SYNC:
         DSL_CPE_Event_S_PmSyncHandle(pContext, &event);
         break;
#endif /* INCLUDE_DSL_CPE_PM_HISTORY*/
#endif /* INCLUDE_DSL_PM*/
      case DSL_EVENT_S_AUTOBOOT_STATUS:
         DSL_CPE_Event_S_AutobootStatusHandle(pContext, &event);
         break;
      case DSL_EVENT_S_SYSTEM_INTERFACE_STATUS:
         DSL_CPE_Event_S_SystemInterfaceStatusHandle(&event);
         break;
      case DSL_EVENT_S_CHANNEL_DATARATE:
         DSL_CPE_Event_S_ChannelDataRateHandle(&event);
         break;
      case DSL_EVENT_I_CHANNEL_DATARATE_SHIFT_THRESHOLD_CROSSING:
         DSL_CPE_Event_S_ChannelDataRateShiftThresholdCrossingHandle(&event);
         break;
      case DSL_EVENT_S_LINIT_FAILURE:
         DSL_CPE_Event_S_LinitFailureHandle(&event);
         break;
      case DSL_EVENT_S_FE_INVENTORY_AVAILABLE:
         DSL_CPE_Event_S_FeInventoryAvailableHandle(pContext, &event);
         break;
      case DSL_EVENT_S_LINE_POWERMANAGEMENT_STATE:
         DSL_CPE_Event_S_LinePowerMgmtStateHandle(&event);
         break;
#if defined(INCLUDE_DSL_CEOC)
      case DSL_EVENT_S_SNMP_MESSAGE_AVAILABLE:
         DSL_CPE_Event_S_SnmpMessageAvailableHandle(&event);
         break;
#endif
#if defined(INCLUDE_DSL_PM)
      #if defined(INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS)
      case DSL_EVENT_I_CHANNEL_THRESHOLD_CROSSING:
         DSL_CPE_Event_I_ChannelThresholdCrossingHandle(&event);
         break;
      #endif /** #if defined(INCLUDE_DSL_CPE_PM_CHANNEL_THRESHOLDS)*/
      case DSL_EVENT_I_LINE_THRESHOLD_CROSSING:
         DSL_CPE_Event_I_LineThresholdCrossingHandle(&event);
         break;
      #if defined(INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS)
      case DSL_EVENT_I_DATA_PATH_THRESHOLD_CROSSING:
         DSL_CPE_Event_I_DataPathThresholdCrossingHandle(&event);
         break;
      #endif /** #if defined(INCLUDE_DSL_CPE_PM_DATA_PATH_THRESHOLDS)*/
#endif /** #if defined(INCLUDE_DSL_PM)*/
      default:
         DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX DSL_CPE_CRLF
            "unhandled event with type: %d" DSL_CPE_CRLF, event.data.nEventType));
         break;
      }
      free(event.data.pData);

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
      /* Do not handle if not string id given */
      if (CLI_EventText[0] != '\0')
      {
         DSL_CPE_CLI_HandleEvent(CLI_EventText);
      }
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   }                            /* for */

   printf ("EventHandler: exit(0)\n");
   return 0;
}

/**
   Start Event handler thread
*/
DSL_Error_t DSL_CPE_EventHandlerStart (
   DSL_CPE_Control_Context_t * pContext
)
{
   if (DSL_CPE_ThreadInit (&pContext->EventControl, "evnthnd", DSL_CPE_EventHandler, DSL_CPE_EVENT_STACKSIZE, DSL_CPE_PRIORITY, (DSL_uint32_t) pContext, 0) == 0)
   {
      return DSL_SUCCESS;
   }

   printf ("Event thread start error\n");

   return DSL_ERROR;
}

DSL_int32_t DSL_CPE_DeviceInit (
   DSL_CPE_Control_Context_t * pContext)
{
   DSL_Error_t nRet = DSL_SUCCESS;
   DSL_uint8_t *pFirmware1 = DSL_NULL;
   DSL_uint32_t nFirmwareSize1 = 0;
   DSL_uint8_t *pFirmware2 = DSL_NULL;
   DSL_uint32_t nFirmwareSize2 = 0;
   DSL_int32_t ret = 0;
   DSL_Init_t init;
   DSL_G997_LineInventoryNeData_t inv;

   memset (&init, 0x00, sizeof (DSL_Init_t));

#ifdef INCLUDE_DSL_CPE_API_VINAX
   /** Just for debug purposes. Further all initialization should be done
       in the gInitCfgData table*/
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_DOWNSTREAM].MinDataRate =
      64000;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_UPSTREAM].MinDataRate = 64000;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_DOWNSTREAM].MaxDataRate =
      140000000;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_UPSTREAM].MaxDataRate =
      140000000;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_DOWNSTREAM].MaxIntDelay = 0;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_UPSTREAM].MaxIntDelay = 0;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_DOWNSTREAM].MinINP = 0;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_UPSTREAM].MinINP = 0;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_DOWNSTREAM].MaxBER =
      DSL_G997_MAX_BER_7;
   gInitCfgData.nDeviceCfg.ChannelConfigData[DSL_UPSTREAM].MaxBER =
      DSL_G997_MAX_BER_7;
/*
   gInitCfgData.nDeviceCfg.LineFeatureData[DSL_DOWNSTREAM].bTrellisEnable =
      DSL_TRUE;
   gInitCfgData.nDeviceCfg.LineFeatureData[DSL_UPSTREAM].bTrellisEnable =
      DSL_TRUE;
   gInitCfgData.nDeviceCfg.LineFeatureData[DSL_DOWNSTREAM].bBitswapEnable =
      DSL_FALSE;
   gInitCfgData.nDeviceCfg.LineFeatureData[DSL_UPSTREAM].bBitswapEnable =
      DSL_FALSE;
*/
#endif

   gInitCfgData.nAutobootStartupMode = DSL_AUTOBOOT_CTRL_START;

   gInitCfgData.nAutobootConfig.nStateMachineOptions.bWaitBeforeLinkActivation =
      g_bWaitBeforeLinkActivation;

   gInitCfgData.nAutobootConfig.nStateMachineOptions.bWaitBeforeConfigWrite =
      g_bWaitBeforeConfigWrite;

   gInitCfgData.nAutobootConfig.nStateMachineOptions.bWaitBeforeRestart =
      g_bWaitBeforeRestart;

   memcpy (&init.data, &gInitCfgData, sizeof (DSL_InitData_t));

   if (bOptimize != 1)
   {
      /* Get Firmware binary 1 */
      if ((g_bFirmware1 != -1) || (strlen(g_sFirmwareName1) > 0))
      {
         nRet = DSL_CPE_LoadFirmwareFromFile(g_sFirmwareName1, &pFirmware1,
            &nFirmwareSize1);
         if (nRet == DSL_SUCCESS)
         {
            if (pFirmware1 != DSL_NULL)
            {
               /* Assign FW binary*/
               init.data.pFirmware = pFirmware1;
               init.data.nFirmwareSize = nFirmwareSize1;
            }
         }
      }

      /* Get Firmware binary 2 */
      if ((g_bFirmware2 != -1) || (strlen(g_sFirmwareName2)))
      {
         nRet = DSL_CPE_LoadFirmwareFromFile(g_sFirmwareName2, &pFirmware2,
            &nFirmwareSize2);
         if (nRet == DSL_SUCCESS)
         {
            if (pFirmware2 != DSL_NULL)
            {
               /* Assign FW binary*/
               init.data.pFirmware2 = pFirmware2;
               init.data.nFirmwareSize2 = nFirmwareSize2;
            }
         }
      }
   }

   if (bXtuOctets == 1)
   {
      /* Use given XTU octets given on init (-i) */
      memcpy (&init.data.nXtseCfg, &g_nXtseInit,
         sizeof(DSL_G997_XTUSystemEnablingData_t));
   }
   else
   {
      /* If no XTU octets are given on init (-i) use device specific default
         configuration */
      memset (&init.data.nXtseCfg.XTSE, 0, sizeof(init.data.nXtseCfg.XTSE));

#ifdef INCLUDE_DSL_CPE_API_VINAX
/*tmp*/
#if 0
      init.data.nXtseCfg.XTSE[0] = 0x17;
      init.data.nXtseCfg.XTSE[1] = 0x01;
      init.data.nXtseCfg.XTSE[2] = 0x14;
      init.data.nXtseCfg.XTSE[3] = 0x01;
      init.data.nXtseCfg.XTSE[4] = 0x40;
      init.data.nXtseCfg.XTSE[5] = 0x05;
      init.data.nXtseCfg.XTSE[6] = 0x04;
#endif
/*tmp*/
      init.data.nXtseCfg.XTSE[7] = 0x7;
#endif
#if defined (INCLUDE_DSL_CPE_API_DANUBE) || defined (INCLUDE_DSL_CPE_API_AMAZON_SE) || \
    defined (INCLUDE_DSL_CPE_API_AMAZON_S)
      init.data.nXtseCfg.XTSE[0] = 0x4;
      init.data.nXtseCfg.XTSE[1] = 0x1;
      init.data.nXtseCfg.XTSE[2] = 0x4;
      init.data.nXtseCfg.XTSE[3] = 0x1;
      init.data.nXtseCfg.XTSE[5] = 0x1;
#endif
   }

   memset(&inv,0x0, sizeof(DSL_G997_LineInventoryNeData_t ));
#ifdef INCLUDE_DSL_CPE_API_VINAX
   memcpy(&inv.Auxiliary.pData, "12344321", 8);
   inv.Auxiliary.nLength = 8;
#endif
   /* $$ ND: this is for testing only --> */
   memcpy (&inv.SerialNumber, "01234567890123456789012345678901", 32);
   memcpy (&inv.SystemVendorID, "23456789", 8);
   memcpy (&inv.VersionNumber, "0123456789012345", 16);
   init.data.pInventory = &inv;
   /* $$ <-- */

   /* Initialize device driver*/
   ret = DSL_CPE_Ioctl (pContext->fd, DSL_FIO_INIT, (DSL_int_t) & init);
   if (ret < 0)
   {
      ret = init.accessCtl.nReturn;

      if (ret == DSL_WRN_ALREADY_INITIALIZED)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX "Already initialized - "
            "ignoring '-i' option!" DSL_CPE_CRLF));
      }
   }

   if (bOptimize == 1)
   {
      if ((g_bFirmware1 != -1) || (strlen(g_sFirmwareName1) > 0))
      {
         ret = DSL_CPE_DownloadFirmware(DSL_FW_REQUEST_ADSL, g_sFirmwareName1, DSL_NULL);
      }
   }

   if (pFirmware1 != DSL_NULL)
   {
      DSL_CPE_Free (pFirmware1);
      pFirmware1 = DSL_NULL;
      nFirmwareSize1 = 0;
   }

   if (pFirmware2 != DSL_NULL)
   {
      DSL_CPE_Free (pFirmware2);
      pFirmware2 = DSL_NULL;
      nFirmwareSize2 = 0;
   }

   return ret;
}


/**
   Termination handler. Will clean up in case of ctrl-c.

   \param sig signal number
*/
DSL_CPE_STATIC  void DSL_CPE_TerminationHandler (
   DSL_int_t sig
)
{
   /* ignore the signal, we'll handle by ourself */
   signal (sig, SIG_IGN);

   if (sig == SIGINT)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX "terminated" DSL_CPE_CRLF));
      DSL_CPE_Termination ();
   }
}

DSL_CPE_STATIC  DSL_void_t DSL_CPE_Termination (
)
{
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_char_t buf[32] = "quit";
#endif

   DSL_CPE_Control_Context_t *pCtrlCtx;

   pCtrlCtx = DSL_CPE_GetGlobalContext();
   if (pCtrlCtx != DSL_NULL)
   {
      pCtrlCtx->bRun = DSL_FALSE;
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   /* quit via the CLI */
   DSL_CPE_CLI_CommandExecute (DSL_CPE_GetGlobalContext ()->fd, buf, DSL_NULL,
      DSL_CPE_STDERR);

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
      "CLI interface of DSL CPE API terminated." DSL_CPE_CRLF));
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */
}

/**
   Control application for the DSL CPE driver.

   \param argc - number of parameters
   \param argv - array of parameter strings

   \return
   DSL_SUCCESS - success
*/
DSL_int_t dsl_cpe_daemon (
   int argc,
   DSL_char_t * argv[])
{
   DSL_int_t ret = 0;
   DSL_char_t device[128] = "";
   DSL_int_t fd = 0, i = 0;
   DSL_AutobootStatus_t status;
   DSL_DBG_ModuleLevel_t nDbgModLvl;
#ifdef VXWORKS
   DSL_int_t stdfds[3];
#endif

   /* main CPE control context */
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   DSL_CPE_Console_Context_t *pConsoleContext = DSL_NULL;
#endif
   DSL_CPE_Control_Context_t nCtrlCtx;
   DSL_CPE_Control_Context_t *pCtrlCtx = &nCtrlCtx;
   DSL_InstanceControl_t instanceConfig;
   DSL_EventStatusMask_t statusEventMaskCfgSet;
   
   memset (pCtrlCtx, 0x00, sizeof (DSL_CPE_Control_Context_t));
   gDSLContext = pCtrlCtx;
   g_nMsgDumpDbgLvl = DSL_DBG_MSG;
#ifdef INCLUDE_DSL_CPE_API_VINAX
   /* Set default Device Number*/
   pCtrlCtx->nDevNum = 0;
#endif /* INCLUDE_DSL_CPE_API_VINAX*/

   /* Initialize firmware files with its default values (might be overwritten
      by user arguments in folowing DSL_CPE_ArgParse() function) */
   g_sFirmwareName1 = DSL_CPE_Malloc(strlen(sDefaultFirmwareName1) + 1);
   g_sFirmwareName2 = DSL_CPE_Malloc(strlen(sDefaultFirmwareName2) + 1);
   if(g_sFirmwareName1)
   {
      strcpy(g_sFirmwareName1, sDefaultFirmwareName1);
   }
   if(g_sFirmwareName2)
   {
      strcpy(g_sFirmwareName2, sDefaultFirmwareName2);
   }

   DSL_CPE_ArgParse (argc, argv);

   /* display version */
   if (bGetVersion != -1)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "DSL CPE Control Application V%s" DSL_CPE_CRLF, &dsl_cpe_ctl_version[4]));

#if 0
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
      ret = DSL_CPE_CLI_CommandExecute (fd, "vig", DSL_NULL, DSL_CPE_STDOUT);
      if (ret >= 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_CRLF));
      }
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */
#endif
      return DSL_SUCCESS;
   }

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   if (g_sAdslScript != DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "Using ADSL autostart script file - %s" DSL_CPE_CRLF, g_sAdslScript));
   }
#ifdef INCLUDE_DSL_CPE_API_VINAX
   if (g_sVdslScript != DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "Using VDSL autostart script file - %s" DSL_CPE_CRLF, g_sVdslScript));
   }
#endif /* INCLUDE_DSL_CPE_API_VINAX */
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */
#ifdef INCLUDE_SCRIPT_NOTIFICATION
   if (g_sRcScript != DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "using script notification file - %s" DSL_CPE_CRLF , g_sRcScript));
   }
#endif /* #ifdef INCLUDE_SCRIPT_NOTIFICATION */

   if (bInit == 1)
   {
      /* Print currently used firmware file(s) */
      if (strlen(g_sFirmwareName1) > 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "using 1st firmware file - %s" DSL_CPE_CRLF , g_sFirmwareName1));
      }
      if (strlen(g_sFirmwareName2) > 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "using 2nd firmware file - %s" DSL_CPE_CRLF , g_sFirmwareName2));
      }
   }

#ifdef INCLUDE_DSL_CPE_API_VINAX
   if (sLowLevCfgName != DSL_NULL)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "using low level configuration file - %s" DSL_CPE_CRLF,
         sLowLevCfgName));

      ret = DSL_CPE_GetInitialLowLevelConfig( sLowLevCfgName,
                                          &(gInitCfgData.nDeviceCfg.cfg) );
      if (ret != DSL_SUCCESS)
     {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "using default low level configuration" DSL_CPE_CRLF));
     }
   }
#endif /* #ifdef INCLUDE_DSL_CPE_API_VINAX*/

   if (bHelp == 1)
   {
      DSL_CPE_Help (argv[0]);
      return DSL_SUCCESS;
   }

#ifdef LINUX
#ifdef USE_DAEMONIZE
   if (bConsole != 1)
   {
      if (daemon (1, bNotSilent) != 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "DSL CPE Control Application cannot daemonize (err=%d)" DSL_CPE_CRLF, errno));
         ret = DSL_ERROR;
         goto DSL_CPE_CONTROL_EXIT;
      }
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "DSL CPE Control Application running in background now!" DSL_CPE_CRLF));
   }
#endif /* USE_DAEMONIZE */
   DSL_CPE_HandlerInstall ();
#endif /* LINUX */

   signal (SIGINT, DSL_CPE_TerminationHandler);

#ifdef INCLUDE_DSL_CPE_API_VINAX
   sprintf (device, "%s/%d", DSL_CPE_DEVICE_NAME, pCtrlCtx->nDevNum);
#else
   sprintf (device, "%s", DSL_CPE_DEVICE_NAME);
#endif /* INCLUDE_DSL_CPE_API_VINAX*/
   
   fd = DSL_CPE_Open (device);
   if (fd < 0)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "cannot open device %s." DSL_CPE_CRLF, device));
      ret = DSL_ERROR;
      goto DSL_CPE_CONTROL_EXIT;
   }

   DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
      "Device %s opened successfully" DSL_CPE_CRLF, device));
   
   /* DSL_CPE_FPrintf(DSL_CPE_STDOUT, "open device: %d %s." DSL_CPE_CRLF, fd, device); */
   pCtrlCtx->fd = fd;

#ifdef DSL_CPE_SOAP_FW_UPDATE
   DSL_CPE_LockCreate(&pCtrlCtx->semFwUpdate);
   pCtrlCtx->firmware.pData = DSL_NULL;
   pCtrlCtx->firmware.nSize = 0;
   pCtrlCtx->firmware2.pData = DSL_NULL;
   pCtrlCtx->firmware2.nSize = 0;
#endif /* DSL_CPE_SOAP_FW_UPDATE */

   pCtrlCtx->bRun = DSL_TRUE;

#ifdef VXWORKS
   /* FIXME: make sure that this console runs on the serial port (fd 3) */
   for (i = 0; i < 3; ++i)
   {
      stdfds[i] = ioTaskStdGet(0,i);
      if (stdfds[i] != ERROR)
         ioTaskStdSet(0,i,3);
   }
#endif

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   if (DSL_CPE_CLI_Init (fd) == DSL_ERROR)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "'DSL_CPE_CLI_Init' failed! Exit dsl_cpe_control..." DSL_CPE_CRLF));
      goto DSL_CPE_CONTROL_EXIT;
   }
#endif

#ifdef DSL_DEBUG_TOOL_INTERFACE
   if (bTcpMessageIntf == 1)
   {
      DSL_CPE_TcpDebugMessageIntfStart(pCtrlCtx);
#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
#ifdef VXWORKS
   #ifdef _MSC_VER
      #pragma message( "WARNING: !!!! TCP messages CLI interface is not fully supported under VxWorks" )
   #else
      #warning "WARNING: !!!! TCP messages CLI interface is not fully supported under VxWorks"
   #endif 
#endif /* VXWORKS*/
      DSL_CPE_TcpDebugCliIntfStart(pCtrlCtx);
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT*/
   }
#endif /* DSL_DEBUG_TOOL_INTERFACE*/

#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
   /* start soap server */
   if (DSL_CPE_SoapInit (fd) < 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "DSL: 'DSL_CPE_SoapInit' failed! Exit dsl_daemon..." DSL_CPE_CRLF));
         goto DSL_CPE_CONTROL_EXIT;
      }
#endif

   /* Configure event and resource handling*/
   instanceConfig.data.bEventActivation = bEventActivation;
   instanceConfig.data.nResourceActivationMask = nResourceActivationMask;
   ret = DSL_CPE_Ioctl (pCtrlCtx->fd, DSL_FIO_INSTANCE_CONTROL_SET,
      (DSL_int_t) &instanceConfig);
   if (ret < 0)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX "Could not configure "
         "event handling (%d)." DSL_CPE_CRLF, instanceConfig.accessCtl.nReturn));
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
   if (bConsole == 1)
   {
      if (DSL_CPE_Console_Init (&pConsoleContext, fd, DSL_CPE_STDIN,
            DSL_CPE_STDOUT) == DSL_ERROR)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "'DSL_CLI_Console_Init' failed! Exit dsl_cpe_control..." DSL_CPE_CRLF));
         goto DSL_CPE_CONTROL_EXIT;
      }
   }
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

   if (bEventActivation)
   {
      /* Unmask (enable) all possible events that are masked (disabled) by
         default after instance event activation. */
      for (i = DSL_EVENT_S_FIRST; i < DSL_EVENT_S_LAST; i++)
      {
         memset(&statusEventMaskCfgSet, 0x00, sizeof(DSL_EventStatusMask_t));
         statusEventMaskCfgSet.data.nEventType = (DSL_EventType_t)i;
         statusEventMaskCfgSet.data.bMask      = DSL_FALSE;
         ret = DSL_CPE_Ioctl(fd, DSL_FIO_EVENT_STATUS_MASK_CONFIG_SET,
            (DSL_int_t) &statusEventMaskCfgSet);
         if (ret < 0)
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX "Could not unmask "
               "event(%d), (%d)!" DSL_CPE_CRLF, i, instanceConfig.accessCtl.nReturn));
         }
      }
   }

   if (bEventActivation == DSL_TRUE)
   {
      DSL_CPE_EventHandlerStart (pCtrlCtx);
   }

   if (bMsgDump == 1)
   {
      nDbgModLvl.data.nDbgModule = DSL_DBG_MESSAGE_DUMP;
      nDbgModLvl.data.nDbgLevel = g_nMsgDumpDbgLvl;
      ret = DSL_CPE_Ioctl (pCtrlCtx->fd, DSL_FIO_DBG_MODULE_LEVEL_SET, (int) &nDbgModLvl);
      if (ret < 0)
   {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "failed to set message dump debug level!" DSL_CPE_CRLF, ret));
      }
   }

   if (bInit == 1)
   {
      ret = DSL_CPE_DeviceInit (pCtrlCtx);
      if (ret < 0)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "Device initialization failed (ret=%d)!" DSL_CPE_CRLF, ret));
      goto DSL_CPE_CONTROL_EXIT;
   }
      else
      {
         {
            DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX
               "Device initialized succeeded (ret=%d)." DSL_CPE_CRLF, ret));
         }
      }
   }

   /* Check current state of autoboot thread and handle FW request if necessary. */
   memset (&status, 0, sizeof(DSL_AutobootStatus_t));
   ret = DSL_CPE_Ioctl (pCtrlCtx->fd, DSL_FIO_AUTOBOOT_STATUS_GET, (DSL_int_t) &status);
   if (ret >= 0)
   {
      if (status.data.nStatus == DSL_AUTOBOOT_STATUS_FW_WAIT)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_WRN, (DSL_CPE_PREFIX "FW reload event "
            "pending - reload FW now." DSL_CPE_CRLF));

         DSL_CPE_Event_S_FirmwareRequestHandle(pCtrlCtx,
            DSL_EVENT_S_FIRMWARE_REQUEST, status.data.nFirmwareRequestType);
      }
   }

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
#ifndef DSL_CPE_REMOVE_PIPE_SUPPORT
   if (DSL_CPE_Pipe_Init (pCtrlCtx) == DSL_ERROR)
   {
      DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
         "'DSL_CPE_Pipe_Init' failed! Exit dsl_cpe_control..." DSL_CPE_CRLF));
      goto DSL_CPE_CONTROL_EXIT;
   }
#endif /* DSL_CPE_REMOVE_PIPE_SUPPORT */

   if (bConsole == 1)
   {
      do
      {
         ret = DSL_CPE_Handle_Console (pConsoleContext);
         if (ret < 0)
         {
            DSL_CPE_Termination ();
            break;
         }
      }
      while (ret != DSL_ERROR);

      ret = DSL_SUCCESS;

      if (DSL_CPE_Console_Shutdown(pConsoleContext) == DSL_ERROR)
      {
         goto DSL_CPE_CONTROL_EXIT;
      }
      pConsoleContext = DSL_NULL;

   }
   else
   {
      dummy_console.bRun = DSL_TRUE;

      DSL_CPE_CLI_Register (&dummy_console.pCLIContext, &dummy_console,
         DSL_CPE_Control_Exit, DSL_NULL);

      while (dummy_console.bRun)
      {
         DSL_CPE_Sleep (1);
      }
   }

   DSL_CPE_CLI_Shutdown(fd);
#else
   while(pCtrlCtx->bRun)
   {
      DSL_CPE_Sleep (1);   
   }
#endif /* INCLUDE_DSL_CPE_CLI_SUPPORT */

DSL_CPE_CONTROL_EXIT:

#ifdef VXWORKS
   /* restore stdio */
   for (i = 0; i < 3; ++i)
   {
      if (stdfds[i] != ERROR)
         ioTaskStdSet(0,i,stdfds[i]);
   }
#endif

   pCtrlCtx->bRun = DSL_FALSE;

   /* wait while CPE Event handler thread ends */
   DSL_CPE_ThreadShutdown (&pCtrlCtx->EventControl, 1000);

   if(fd >= 0)
   {
      /* close device */
      ret = (DSL_int32_t) (DSL_CPE_Close (fd));
      if (ret < 0)
      {
         DSL_CCA_DEBUG(DSL_CCA_DBG_ERR, (DSL_CPE_PREFIX
            "error on closing device %s." DSL_CPE_CRLF, device));
      }
   }

#ifdef INCLUDE_SCRIPT_NOTIFICATION
   if(g_sRcScript != DSL_NULL)
   {
      DSL_CPE_Free (g_sRcScript);
      g_sRcScript = DSL_NULL;
   }
#endif

   if(g_sFirmwareName1 != DSL_NULL)
   {
      DSL_CPE_Free (g_sFirmwareName1);
      g_sFirmwareName1 = DSL_NULL;
   }

   if(g_sFirmwareName2 != DSL_NULL)
   {
      DSL_CPE_Free (g_sFirmwareName2);
      g_sFirmwareName2 = DSL_NULL;
   }

#ifdef INCLUDE_DSL_CPE_SOAP_SUPPORT
   if (sSoapRemoteServer)
   {
      DSL_CPE_Free (sSoapRemoteServer);
      sSoapRemoteServer = DSL_NULL;
   }
#endif

#ifdef DSL_CPE_SOAP_FW_UPDATE
   if(pCtrlCtx->firmware.pData != DSL_NULL)
   {
      DSL_CPE_Free (pCtrlCtx->firmware.pData);
      pCtrlCtx->firmware.pData = DSL_NULL;
      pCtrlCtx->firmware.nSize = 0;
   }

   if(pCtrlCtx->firmware2.pData != DSL_NULL)
   {
      DSL_CPE_Free (pCtrlCtx->firmware2.pData);
      pCtrlCtx->firmware2.pData = DSL_NULL;
      pCtrlCtx->firmware2.nSize = 0;
   }
#endif

#ifdef DSL_DEBUG_TOOL_INTERFACE
   if(sTcpMessagesSocketAddr != DSL_NULL)
   {
      DSL_CPE_Free (sTcpMessagesSocketAddr);
      sTcpMessagesSocketAddr = DSL_NULL;
   }
#endif

#ifdef INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT
#ifdef INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT
   if(g_sAdslScript != DSL_NULL)
   {
      DSL_CPE_Free (g_sAdslScript);
      g_sAdslScript = DSL_NULL;
   }

   if(g_sVdslScript != DSL_NULL)
   {
      DSL_CPE_Free (g_sVdslScript);
      g_sVdslScript = DSL_NULL;
   }
#endif /* INCLUDE_DSL_CPE_CMV_SCRIPTS_SUPPORT */
#endif /* INCLUDE_DSL_CPE_FILESYSTEM_SUPPORT */

   return ret;
}

#if defined(LINUX) || defined(WIN32)
int main (int argc, char *argv[])
{
#ifdef DSL_CPE_SIMULATOR
   DSL_SimulatorInitialize();
#endif

   return dsl_cpe_daemon (argc, argv);
}
#endif

#ifdef INCLUDE_DSL_CPE_CLI_SUPPORT
DSL_CPE_STATIC  DSL_Error_t DSL_CPE_Control_Exit (DSL_void_t * pContext)
{
   dummy_console_t *pConsole = pContext;
   pConsole->bRun = DSL_FALSE;
   return DSL_SUCCESS;
}
#endif

/**
sscanf implementation with uint8 support.
*/
DSL_int32_t DSL_CPE_sscanf (
   DSL_char_t *buf,
   DSL_char_t const *fmt,
   ...
)
{
#ifndef _lint
   va_list marker;
#endif
   DSL_char_t const *p, *a;
   DSL_char_t *r = 0, *s = buf;
   DSL_char_t *v8 = 0;
   DSL_int_t ret = 0, mode = 32, base = 10, array = 0;
   DSL_int_t i;
   DSL_int_t *v32 = 0;
   short *v16 = 0;
   unsigned int *vu32 = 0;
   unsigned short *vu16 = 0;
   unsigned char *vu8 = 0;

   if ((buf == DSL_NULL) || (fmt == DSL_NULL))
      return ret;

#ifndef _lint
   va_start (marker, fmt);
#endif

   if (s == DSL_NULL)
      goto SF_END;

   for (p = fmt; *p; p++)
   {
      if (*p != '%')
         continue;

      if (s == DSL_NULL)
         goto SF_END;

      /* skip spaces and tabs */
      while ((*s == ' ') || (*s == '\t'))
         s++;

      if (*s == 0)
         goto SF_END;

      switch (*++p)
      {
      case 0:
         /* ret = 0; */
         goto SF_END;

         /* 8 bit */
      case 'b':
         mode = 8;
         p++;
         break;

         /* 16 bit */
      case 'h':
         mode = 16;
         p++;
         break;

         /* 32 bit */
      default:
         mode = 32;
         break;
      }

      switch (*p)
      {
      case 'd':
      case 'i':
      case 'u':
         base = 10;
         break;

      case 'x':
         base = 16;
         break;

      default:
         break;
      }

      a = p + 1;
      i = 0;

      array = 1;

      switch (*a)
      {
      case '[':
         a++;
         if (*a)
         {
            array = (DSL_char_t) strtol (a, DSL_NULL, 10);
            do
            {
               a++;
               if (*a == ']')
               {
                  a++;
                  break;
               }
            } while (*a);
         }
         break;

      default:
         break;
      }

      switch (*p)
      {
      case 0:
         /* ret = 0; */
         goto SF_END;

         /* string */
      case 's':
         {
#ifndef _lint
            r = va_arg (marker, DSL_char_t *);
#endif
            if (r != DSL_NULL)
            {
               DSL_char_t *q = s;

               if (q)
               {
                  do
                  {
                     if ((*q == ' ') || (*q == '\t'))
                        q++;
                     else
                        break;
                  } while (*q);

                  if (*q)
                  {
                     do
                     {
                        if ((*q != ' ') && (*q != '\t') && (*q != '\n') &&
                           (*q != '\r'))
                           *r++ = *q++;
                        else
                           break;
                     } while (*q);
                     s = q;
                     *r = 0;
                     ret++;
                  }
               }
            }
            break;
         }
         /* signed */
      case 'd':
      case 'i':
         {
            switch (mode)
            {
            case 8:
#ifndef _lint
               v8 = va_arg (marker, DSL_char_t *);
#endif
               if (v8 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        v8[i] = (DSL_char_t) strtol (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            case 16:
#ifndef _lint
               v16 = va_arg (marker, short *);
#endif
               if (v16 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        v16[i] = (short) strtol (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            case 32:
#ifndef _lint
               v32 = va_arg (marker, DSL_int_t *);
#endif
               if (v32 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        v32[i] = (DSL_int_t) strtol (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            default:
               break;
            }
            break;
         }

         /* unsigned */
      case 'u':
         /* hexadecimal */
      case 'x':
         {
            switch (mode)
            {
            case 8:
#ifndef _lint
               vu8 = va_arg (marker, unsigned char *);
#endif
               if (vu8 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        vu8[i] = (unsigned char) strtoul (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            case 16:
#ifndef _lint
               vu16 = va_arg (marker, unsigned short *);
#endif
               if (vu16 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        vu16[i] = (unsigned short) strtoul (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            case 32:
#ifndef _lint
               vu32 = va_arg (marker, unsigned int *);
#endif
               if (vu32 != DSL_NULL)
               {
                  for (i = 0; i < array; i++)
                  {
                     if (s && *s != 0)
                     {
                        vu32[i] = (unsigned int) strtoul (s, &s, base);
                     }
                     else
                     {
                        break;
                     }
                  }
                  if (i == array)
                     ret++;
               }
               break;

            default:
               break;
            }
            break;

      default:
            break;
         }
      }

      if (array > 1)
      {
         p = a;
      }
   }

 SF_END:

#ifndef _lint
   va_end (marker);
#endif

   return ret;
}
