/*
 *  Gary Jennejohn (C) 2003 <gj@denx.de>
 *
 *  This program is free software; you can distribute it and/or modify it
 *  under the terms of the GNU General Public License (Version 2) as
 *  published by the Free Software Foundation.
 *
 *  This program is distributed in the hope it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 * Routines for generic manipulation of the interrupts found on the 
 * AMAZON_SE boards.
 *
 * HISTORY
 * $Version $Date      $Author     $Comment
 *  2.0.0   27/04/07   Teh Kok How Use unified function calls and defines
 *  				   between Danube and Amazon_SE. Benefits is,
 *			   	   of course, better code management.
 */
#include <linux/config.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/kernel_stat.h>
#include <linux/module.h>

#include <asm/bootinfo.h>
#include <asm/irq.h>
#ifdef CONFIG_AMAZON_SE
#include <asm/amazon_se/amazon_se.h>
#include <asm/amazon_se/irq.h>
#elif defined(CONFIG_DANUBE)
#include <asm/danube/danube.h>
#include <asm/danube/irq.h>
#else
#error Platform Undefined!!!
#endif
#include <asm/traps.h>
#ifdef CONFIG_KGDB
#include <asm/gdb-stub.h>
#endif

extern asmlinkage void mipsIRQ (void);
extern asmlinkage void mipsIRQ0 (void);
extern asmlinkage void mipsIRQ1 (void);
extern asmlinkage void mipsIRQ2 (void);
extern asmlinkage void mipsIRQ3 (void);
extern asmlinkage void mipsIRQ4 (void);
extern asmlinkage void mipsIRQ5 (void);
extern asmlinkage void mipsIRQ6 (void);
extern asmlinkage void mipsIRQ7 (void);
extern asmlinkage void end_mipsIRQ0(void);
extern asmlinkage void end_mipsIRQ1(void);
extern asmlinkage void end_mipsIRQ2(void);
extern asmlinkage void end_mipsIRQ3(void);
extern asmlinkage void end_mipsIRQ4(void);
extern asmlinkage void end_mipsIRQ5(void);
extern asmlinkage void end_mipsIRQ6(void);
extern asmlinkage void end_mipsIRQ7(void);

//extern asmlinkage void mipsIRQ(void);
#if 0
extern void do_IRQ(int irq, struct pt_regs *regs);
#endif
#undef BSP_INTERRUPT_DEBUG_MSG
#ifdef BSP_INTERRUPT_DEBUG_MSG
#define BSP_INTERRUPT_DMSG(fmt,args...) printk(KERN_INFO "[%s %d]: " fmt, __FUNCTION__ , __LINE__, ##args)
#else
#define BSP_INTERRUPT_DMSG(x...)
#endif 

#ifdef CONFIG_DANUBE

#define BSP_PLATFORM		"DANUBE"
#define BSP_DISABLE_IRQ 	disable_danube_irq
#define BSP_MASK_AND_ACK_IRQ	mask_and_ack_danube_irq
#define BSP_ENABLE_IRQ		enable_danube_irq
#define BSP_STARTUP_IRQ		startup_danube_irq
#define BSP_END_IRQ		end_danube_irq

#define BSP_ICU_IM0_IER		DANUBE_ICU_IM0_IER
#define BSP_ICU_IM1_IER		DANUBE_ICU_IM1_IER
#define BSP_ICU_IM2_IER		DANUBE_ICU_IM2_IER
#define BSP_ICU_IM3_IER		DANUBE_ICU_IM3_IER
#define BSP_ICU_IM4_IER		DANUBE_ICU_IM4_IER

#define BSP_ICU_IM0_IER_IR	DANUBE_ICU_IM0_IER_IR
#define BSP_ICU_IM1_IER_IR	DANUBE_ICU_IM1_IER_IR
#define BSP_ICU_IM2_IER_IR	DANUBE_ICU_IM2_IER_IR
#define BSP_ICU_IM3_IER_IR	DANUBE_ICU_IM3_IER_IR
#define BSP_ICU_IM4_IER_IR	DANUBE_ICU_IM4_IER_IR

#define BSP_ICU_IM0_ISR		DANUBE_ICU_IM0_ISR
#define BSP_ICU_IM1_ISR		DANUBE_ICU_IM1_ISR
#define BSP_ICU_IM2_ISR		DANUBE_ICU_IM2_ISR
#define BSP_ICU_IM3_ISR		DANUBE_ICU_IM3_ISR
#define BSP_ICU_IM4_ISR		DANUBE_ICU_IM4_ISR

#define BSP_ICU_IM0_ISR_IR	DANUBE_ICU_IM0_ISR_IR
#define BSP_ICU_IM1_ISR_IR	DANUBE_ICU_IM1_ISR_IR
#define BSP_ICU_IM2_ISR_IR	DANUBE_ICU_IM2_ISR_IR
#define BSP_ICU_IM3_ISR_IR	DANUBE_ICU_IM3_ISR_IR
#define BSP_ICU_IM4_ISR_IR	DANUBE_ICU_IM4_ISR_IR

#define BSP_ICU_IM0_IOSR	DANUBE_ICU_IM0_IOSR
#define BSP_ICU_IM1_IOSR	DANUBE_ICU_IM1_IOSR
#define BSP_ICU_IM2_IOSR	DANUBE_ICU_IM2_IOSR
#define BSP_ICU_IM3_IOSR	DANUBE_ICU_IM3_IOSR
#define BSP_ICU_IM4_IOSR	DANUBE_ICU_IM4_IOSR

#define BSP_ICU_IM0_IOSR	DANUBE_ICU_IM0_IOSR
#define BSP_ICU_IM1_IOSR	DANUBE_ICU_IM1_IOSR
#define BSP_ICU_IM2_IOSR	DANUBE_ICU_IM2_IOSR
#define BSP_ICU_IM3_IOSR	DANUBE_ICU_IM3_IOSR
#define BSP_ICU_IM4_IOSR	DANUBE_ICU_IM4_IOSR

#define BSP_ICU_IM_VEC		DANUBE_ICU_IM_VEC
#define BSP_ICU_IM0_VEC_MASK	DANUBE_ICU_IM0_VEC_MASK
#define BSP_ICU_IM4_VEC_MASK	DANUBE_ICU_IM4_VEC_MASK

#define BSP_EBU_PCC_ISTAT	DANUBE_EBU_PCC_ISTAT

#elif defined(CONFIG_AMAZON_SE)

#define BSP_PLATFORM		"AMAZON_SE"
#define BSP_DISABLE_IRQ 	disable_amazon_se_irq
#define BSP_MASK_AND_ACK_IRQ	mask_and_ack_amazon_se_irq
#define BSP_ENABLE_IRQ		enable_amazon_se_irq
#define BSP_STARTUP_IRQ		startup_amazon_se_irq
#define BSP_END_IRQ		end_amazon_se_irq

#define BSP_ICU_IM0_IER		AMAZON_SE_ICU_IM0_IER
#define BSP_ICU_IM1_IER		AMAZON_SE_ICU_IM1_IER
#define BSP_ICU_IM2_IER		AMAZON_SE_ICU_IM2_IER
#define BSP_ICU_IM3_IER		AMAZON_SE_ICU_IM3_IER
#define BSP_ICU_IM4_IER		AMAZON_SE_ICU_IM4_IER

#define BSP_ICU_IM0_IER_IR	AMAZON_SE_ICU_IM0_IER_IR
#define BSP_ICU_IM1_IER_IR	AMAZON_SE_ICU_IM1_IER_IR
#define BSP_ICU_IM2_IER_IR	AMAZON_SE_ICU_IM2_IER_IR
#define BSP_ICU_IM3_IER_IR	AMAZON_SE_ICU_IM3_IER_IR
#define BSP_ICU_IM4_IER_IR	AMAZON_SE_ICU_IM4_IER_IR

#define BSP_ICU_IM0_ISR		AMAZON_SE_ICU_IM0_ISR
#define BSP_ICU_IM1_ISR		AMAZON_SE_ICU_IM1_ISR
#define BSP_ICU_IM2_ISR		AMAZON_SE_ICU_IM2_ISR
#define BSP_ICU_IM3_ISR		AMAZON_SE_ICU_IM3_ISR
#define BSP_ICU_IM4_ISR		AMAZON_SE_ICU_IM4_ISR

#define BSP_ICU_IM0_ISR_IR	AMAZON_SE_ICU_IM0_ISR_IR
#define BSP_ICU_IM1_ISR_IR	AMAZON_SE_ICU_IM1_ISR_IR
#define BSP_ICU_IM2_ISR_IR	AMAZON_SE_ICU_IM2_ISR_IR
#define BSP_ICU_IM3_ISR_IR	AMAZON_SE_ICU_IM3_ISR_IR
#define BSP_ICU_IM4_ISR_IR	AMAZON_SE_ICU_IM4_ISR_IR

#define BSP_ICU_IM0_IOSR	AMAZON_SE_ICU_IM0_IOSR
#define BSP_ICU_IM1_IOSR	AMAZON_SE_ICU_IM1_IOSR
#define BSP_ICU_IM2_IOSR	AMAZON_SE_ICU_IM2_IOSR
#define BSP_ICU_IM3_IOSR	AMAZON_SE_ICU_IM3_IOSR
#define BSP_ICU_IM4_IOSR	AMAZON_SE_ICU_IM4_IOSR

#define BSP_ICU_IM0_IOSR	AMAZON_SE_ICU_IM0_IOSR
#define BSP_ICU_IM1_IOSR	AMAZON_SE_ICU_IM1_IOSR
#define BSP_ICU_IM2_IOSR	AMAZON_SE_ICU_IM2_IOSR
#define BSP_ICU_IM3_IOSR	AMAZON_SE_ICU_IM3_IOSR
#define BSP_ICU_IM4_IOSR	AMAZON_SE_ICU_IM4_IOSR

#define BSP_ICU_IM_VEC		AMAZON_SE_ICU_IM_VEC
#define BSP_ICU_IM0_VEC_MASK	AMAZON_SE_ICU_IM0_VEC_MASK
#define BSP_ICU_IM4_VEC_MASK	AMAZON_SE_ICU_IM4_VEC_MASK

#define BSP_EBU_PCC_ISTAT	AMAZON_SE_EBU_PCC_ISTAT

#endif
u32 hardware_irq_count[NR_CPUS][6];

void BSP_DISABLE_IRQ(unsigned int irq_nr)
{
	BSP_INTERRUPT_DMSG("irq: %d\n", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
		/* access IM0 */
		*BSP_ICU_IM0_IER &= (~BSP_ICU_IM0_IER_IR(irq_nr));
	else if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
		/* access IM1 */
		*BSP_ICU_IM1_IER &= (~BSP_ICU_IM1_IER_IR(irq_nr - 32));
	else if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
		/* access IM2 */
		*BSP_ICU_IM2_IER &= (~BSP_ICU_IM2_IER_IR(irq_nr - 64));
	else if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
		/* access IM3 */
		*BSP_ICU_IM3_IER &= (~BSP_ICU_IM3_IER_IR((irq_nr - 96)));
	else if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
		/* access IM4 */
		*BSP_ICU_IM4_IER &= (~BSP_ICU_IM4_IER_IR((irq_nr - 128)));
}
EXPORT_SYMBOL(BSP_DISABLE_IRQ);
void BSP_MASK_AND_ACK_IRQ(unsigned int irq_nr)
{
	BSP_INTERRUPT_DMSG("irq: %d\n", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0) {
		/* access IM0 */
		/* mask */
		*BSP_ICU_IM0_IER &= ~BSP_ICU_IM0_IER_IR(irq_nr);
		/* ack */
		*BSP_ICU_IM0_ISR = BSP_ICU_IM0_ISR_IR(irq_nr);
	} else if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0) {
		/* access IM1 */
		/* mask */
		*BSP_ICU_IM1_IER &= ~BSP_ICU_IM1_IER_IR(irq_nr - 32);
		/* ack */
		*BSP_ICU_IM1_ISR = BSP_ICU_IM1_ISR_IR(irq_nr - 32);
	} else if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0) {
		/* access IM2 */
		/* mask */
		*BSP_ICU_IM2_IER &= ~BSP_ICU_IM2_IER_IR(irq_nr - 64);
		/* ack */
		*BSP_ICU_IM2_ISR = BSP_ICU_IM2_ISR_IR(irq_nr - 64);
	} else if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0) {
		/* access IM3 */
		/* mask */
		*BSP_ICU_IM3_IER &= ~BSP_ICU_IM3_IER_IR(irq_nr - 96);
		/* ack */
		*BSP_ICU_IM3_ISR = BSP_ICU_IM3_ISR_IR(irq_nr - 96);
	} else if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0) {
		/* access IM4 */
		/* mask */		
		*BSP_ICU_IM4_IER &= ~BSP_ICU_IM4_IER_IR(irq_nr - 128);
		/* ack */
		*BSP_ICU_IM4_ISR = BSP_ICU_IM4_ISR_IR(irq_nr - 128);
	}
}
EXPORT_SYMBOL(BSP_MASK_AND_ACK_IRQ);
void BSP_ENABLE_IRQ(unsigned int irq_nr)
{
	BSP_INTERRUPT_DMSG("irq: %d\n", irq_nr);
	/* have to access the correct register here */
	if (irq_nr <= INT_NUM_IM0_IRL31 && irq_nr >= INT_NUM_IM0_IRL0)
		/* access IM0 */
		*BSP_ICU_IM0_IER |= BSP_ICU_IM0_IER_IR(irq_nr);
	else if (irq_nr <= INT_NUM_IM1_IRL31 && irq_nr >= INT_NUM_IM1_IRL0)
		/* access IM1 */
		*BSP_ICU_IM1_IER |= BSP_ICU_IM1_IER_IR(irq_nr - 32);
	else if (irq_nr <= INT_NUM_IM2_IRL31 && irq_nr >= INT_NUM_IM2_IRL0)
		/* access IM2 */
		*BSP_ICU_IM2_IER |= BSP_ICU_IM2_IER_IR(irq_nr - 64);
	else if (irq_nr <= INT_NUM_IM3_IRL31 && irq_nr >= INT_NUM_IM3_IRL0)
		/* access IM3 */
		*BSP_ICU_IM3_IER |= BSP_ICU_IM3_IER_IR((irq_nr - 96));
	else if (irq_nr <= INT_NUM_IM4_IRL31 && irq_nr >= INT_NUM_IM4_IRL0)
		/* access IM4 */
		*BSP_ICU_IM4_IER |= BSP_ICU_IM4_IER_IR((irq_nr - 128));
}
EXPORT_SYMBOL(BSP_ENABLE_IRQ);

static unsigned int BSP_STARTUP_IRQ(unsigned int irq)
{
	BSP_ENABLE_IRQ(irq);
	return 0; /* never anything pending */
}

#define BSP_SHUTDOWN_IRQ BSP_DISABLE_IRQ

static void BSP_END_IRQ(unsigned int irq)
{
	if (!(irq_desc[irq].status & (IRQ_DISABLED|IRQ_INPROGRESS)))
		BSP_ENABLE_IRQ(irq);
}

static struct hw_interrupt_type bsp_irq_type = {
	BSP_PLATFORM,
	BSP_STARTUP_IRQ,
	BSP_SHUTDOWN_IRQ,
	BSP_ENABLE_IRQ,
	BSP_DISABLE_IRQ,
	BSP_MASK_AND_ACK_IRQ,
	BSP_END_IRQ,
	NULL
};
static unsigned int startup_none(unsigned int irq) { return 0; }
static void enable_none(unsigned int irq) { }
static void disable_none(unsigned int irq) { }
static void ack_none(unsigned int irq) { }

/* startup is the same as "enable", shutdown is same as "disable" */
#define shutdown_none   disable_none
#define end_none        enable_none


struct hw_interrupt_type bsp_timer_irq_type = {
        "TIMER",
        startup_none,
        shutdown_none,
        enable_none,
        disable_none,
        ack_none,
        end_none
};

#define IVEC_BUG_FIX
#ifdef IVEC_BUG_FIX
/**
 * Find least significant bit.
 * This function searches for the least significant bit in 32bit value.
 *
 * \param  x - The value to be checked.
 * \return Position of the least significant bit (0..31)
 */
static inline int ls1bit32(unsigned int x)
{
        register int i;
        for (i = 0; i < 32; i++) {
                if (x & (1 << i))
                        break;
        }
        return i;

}
#endif /* IVEC_BUG_FIX */
/* Cascaded interrupts from IM0 */
void hw0_irqdispatch(struct pt_regs *regs)
{
	u32 irq;
	int cpu = smp_processor_id();

#ifdef IVEC_BUG_FIX
	irq = (*BSP_ICU_IM0_IOSR); 
	/* if int_status == 0, then the interrupt has already been cleared */
	if (irq == 0)
		return;
	irq = ls1bit32(irq);
#else
	irq = (*BSP_ICU_IM_VEC) & BSP_ICU_IM0_VEC_MASK;
	if (irq == 0) 
		return;
	irq--;
#endif /*IVEC_BUG_FIX*/
	if (irq == 22){
		/*clear EBU interrupt */
		*(BSP_EBU_PCC_ISTAT) |= 0x10;
	}
	BSP_INTERRUPT_DMSG("irq: %d\n", irq);

	hardware_irq_count[cpu][0]++;
	do_IRQ((int)irq, regs);

	return;
}

/* Cascaded interrupts from IM1 */
void hw1_irqdispatch(struct pt_regs *regs)
{
	u32 irq;
	int cpu = smp_processor_id();

#ifdef IVEC_BUG_FIX
	irq = (*BSP_ICU_IM1_IOSR); 
	/* if int_status == 0, then the interrupt has already been cleared */
	if (irq == 0)
		return;
	irq = ls1bit32(irq)+32;
#else
	irq = ((*BSP_ICU_IM_VEC) & BSP_ICU_IM4_VEC_MASK) >>6;
	if (irq == 0) 
		return;
	irq+=31;
#endif /*IVEC_BUG_FIX*/
	BSP_INTERRUPT_DMSG("irq: %d\n", irq);

	hardware_irq_count[cpu][1]++;
	do_IRQ((int)irq, regs);

	return;
}
/* Cascaded interrupts from IM2 */
void hw2_irqdispatch(struct pt_regs *regs)
{
	u32 irq;
	int cpu = smp_processor_id();

#ifdef IVEC_BUG_FIX
	irq = (*BSP_ICU_IM2_IOSR); 
	/* if int_status == 0, then the interrupt has already been cleared */
	if (irq == 0)
		return;
	irq = ls1bit32(irq)+64;
#else
	irq = ((*BSP_ICU_IM_VEC) & BSP_ICU_IM4_VEC_MASK) >>12;
	if (irq == 0) 
		return;
	irq += 63;
#endif /*IVEC_BUG_FIX*/
	BSP_INTERRUPT_DMSG("irq: %d\n", irq);

	hardware_irq_count[cpu][2]++;
	do_IRQ((int)irq, regs);

	return;
}
/* Cascaded interrupts from IM3 */
void hw3_irqdispatch(struct pt_regs *regs)
{
	u32 irq;
	int cpu = smp_processor_id();

#ifdef IVEC_BUG_FIX
	irq = (*BSP_ICU_IM3_IOSR); 
	/* if int_status == 0, then the interrupt has already been cleared */
	if (irq == 0)
		return;
	irq = ls1bit32(irq)+96;
#else
	irq = ((*BSP_ICU_IM_VEC) & BSP_ICU_IM4_VEC_MASK) >>18;
	if (irq ==0) 
		return;
	irq += 95;
#endif /*IVEC_BUG_FIX*/
	BSP_INTERRUPT_DMSG("irq: %d\n", irq);

	hardware_irq_count[cpu][3]++;
	do_IRQ((int)irq, regs);

	return;
}
/* Cascaded interrupts from IM4 */
void hw4_irqdispatch(struct pt_regs *regs)
{
	u32 irq;
	int cpu = smp_processor_id();

#ifdef IVEC_BUG_FIX
	irq = (*BSP_ICU_IM4_IOSR); 
	/* if int_status == 0, then the interrupt has already been cleared */
	if (irq == 0)
		return;
	irq = ls1bit32(irq)+128;
#else
	irq = ((*BSP_ICU_IM_VEC) & BSP_ICU_IM4_VEC_MASK) >> 24;
	if (irq == 0) 
		return;
	irq += 127;
#endif /*IVEC_BUG_FIX*/
	BSP_INTERRUPT_DMSG("irq: %d\n", irq);

	hardware_irq_count[cpu][4]++;
	do_IRQ((int)irq, regs);

	return;
}
/* Cascaded interrupts from IM5 */
void
hw5_irqdispatch (struct pt_regs *regs)
{
        int cpu = smp_processor_id();
        hardware_irq_count[cpu][5]++;
        do_IRQ(MIPS_CPU_TIMER_IRQ, regs);
        return;
}

int bus_error_handler(struct pt_regs *regs, int is_fixup)
{
	/*TODO: bus error */
	printk("TODO: BUS error\n");
	return MIPS_BE_FATAL;
}


/* Function for careful CP0 interrupt mask access */
#define COMPAT_IRQ
#ifndef COMPAT_IRQ
#define INTCTL_VS INTCTL_VS512
#endif
void __init init_IRQ(void)
{
	int i;

	BSP_INTERRUPT_DMSG("\n");

	board_be_handler = &bus_error_handler;

	init_generic_irq();

	/* mask all interrupt sources */
	*BSP_ICU_IM0_IER = 0;
	*BSP_ICU_IM1_IER = 0;
	*BSP_ICU_IM2_IER = 0;
	*BSP_ICU_IM3_IER = 0;
	*BSP_ICU_IM4_IER = 0;

	/* Now safe to set the exception vector. */
	#ifdef COMPAT_IRQ
	set_except_vector(0, mipsIRQ);
	#else

        /* Setting up Vectored Interrupt Table
	* Vectored Interrupt Mode is used when the following conditions are met:
	* CONFIG3[VInt] = 1
	* CONFIG3[Veic] = 0
	* INTCTL[VS]   != 0
	* CAUSE[IV]     = 1
	* STATUS[BEV]   = 0
	* A vector number is generated by the interrupt control logic. This number is combined with IntCtlVS to create
	* interrupt offset, which is added to 0x200 to create exception vector offset.
	* Vector Number 0 and 1 = Software interrupt 0 and 1
	* Vector Number 2 to 7  = Hardware interrupt 0 to 5
	*/
        if ((end_mipsIRQ0 - mipsIRQ0) >= INTCTL_VS ||
            (end_mipsIRQ1 - mipsIRQ1) >= INTCTL_VS ||
            (end_mipsIRQ2 - mipsIRQ2) >= INTCTL_VS ||
            (end_mipsIRQ3 - mipsIRQ3) >= INTCTL_VS ||
            (end_mipsIRQ4 - mipsIRQ4) >= INTCTL_VS ||
            (end_mipsIRQ5 - mipsIRQ5) >= INTCTL_VS ||
            (end_mipsIRQ6 - mipsIRQ6) >= INTCTL_VS ||
            (end_mipsIRQ7 - mipsIRQ7) >= INTCTL_VS )
            panic ("%s %s %d: Interrupt Vector spacing >= %d bytes!\n", __FILE__, __func__, __LINE__,INTCTL_VS);
        #if 0
        /* *EITHER* do it here or do it in ld.script */
        memset((void*)(KSEG1 + 0x200), 0, 0x200 * 8);
        memcpy((void*)(KSEG1 + 0x200), mipsIRQ0, end_mipsIRQ0 - mipsIRQ0);
        memcpy((void*)(KSEG1 + 0x400), mipsIRQ1, end_mipsIRQ1 - mipsIRQ1);
        memcpy((void*)(KSEG1 + 0x600), mipsIRQ2, end_mipsIRQ2 - mipsIRQ2);
        memcpy((void*)(KSEG1 + 0x800), mipsIRQ3, end_mipsIRQ3 - mipsIRQ3);
        memcpy((void*)(KSEG1 + 0xA00), mipsIRQ4, end_mipsIRQ4 - mipsIRQ4);
        memcpy((void*)(KSEG1 + 0xC00), mipsIRQ5, end_mipsIRQ5 - mipsIRQ5);
        memcpy((void*)(KSEG1 + 0xE00), mipsIRQ6, end_mipsIRQ6 - mipsIRQ6);
        memcpy((void*)(KSEG1 + 0x1000), mipsIRQ7, end_mipsIRQ7 - mipsIRQ7);
        #endif
        write_c0_intctl(INTCTL_VS); /* This will enable vectored interrupt mode by setting INTCTL[VS] != 0 */

        printk(KERN_INFO "CONFIG  = %#08x\n", read_c0_config());
        printk(KERN_INFO "CONFIG1 = %#08x\n", read_c0_config1());
        printk(KERN_INFO "CONFIG2 = %#08x\n", read_c0_config2());
        printk(KERN_INFO "CONFIG3 = %#08x\n", read_c0_config3());
        printk(KERN_INFO "STATUS  = %#08x\n", read_c0_status());
        printk(KERN_INFO "CAUSE   = %#08x\n", read_c0_cause());
        printk(KERN_INFO "INTCTL  = %#08x\n", read_c0_intctl());
        printk(KERN_INFO "SRSCTL  = %#08x\n", read_c0_srsctl());
        printk(KERN_INFO "SRSMAP  = %#08x\n", read_c0_srsmap());

        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0x200, *(u32*)(KSEG0 + 0x200));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0x400, *(u32*)(KSEG0 + 0x400));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0x600, *(u32*)(KSEG0 + 0x600));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0x800, *(u32*)(KSEG0 + 0x800));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0xA00, *(u32*)(KSEG0 + 0xA00));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0xC00, *(u32*)(KSEG0 + 0xC00));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0xE00, *(u32*)(KSEG0 + 0xE00));
        printk(KERN_INFO "%#x: %#x\n", KSEG0 + 0x1000, *(u32*)(KSEG0 + 0x1000));
        #endif

	for (i = 0; i <= INT_NUM_IM4_IRL31; i++) {
		irq_desc[i].status	= IRQ_DISABLED;
		irq_desc[i].action	= 0;
		irq_desc[i].depth	= 1;
		irq_desc[i].handler	= &bsp_irq_type;
	}
        /* This entry is for timer interrupt */
        /* /proc/irq/ needs irq_desc[].handler to create an entry for every interrupt */
        irq_desc[MIPS_CPU_TIMER_IRQ].status = IRQ_PER_CPU;
        irq_desc[MIPS_CPU_TIMER_IRQ].action = 0;
        irq_desc[MIPS_CPU_TIMER_IRQ].depth = 1;
        irq_desc[MIPS_CPU_TIMER_IRQ].handler = &bsp_timer_irq_type;

	set_c0_status(IE_IRQ0 | IE_IRQ1 | IE_IRQ2 | IE_IRQ3 | IE_IRQ4 | IE_IRQ5);

#ifdef CONFIG_KGDB
	set_debug_traps();
	breakpoint();
#endif
}
